"""
.. highlight:: console

============================
The CLI interface of Pharmpy
============================

Examples
--------

Example argument lines (only preceeded by ``pharmpy`` or ``python3 -m pharmpy``)::

    # filter data (as model prescribes) and output to file 'output.csv'
    pharmpy data write run1.mod -o output.csv

Entrypoints and Subcommands
---------------------------

Standard practice of packaging is to define "entrypoints" in the setuptools ``setup.py``. Currently:

.. code-block:: python

  entry_points={
      'console_scripts': [
          'pharmpy           = pharmpy.cli:main',
      ]
  },

Main entrypoint is ``pharmpy`` (which maps :func:`main`), but each subcommand could be given a
shortcutting entrypoint (e.g. :func:`sumo`). This would be accomplished via invoking:

.. code-block:: python

    CLI(*args_of_sumo, subcommand='sumo')

to skip straight to subcommand parser.

Installing Pharmpy Wheel-build should install all of these to PATH. Then, these are clean targets
for post-install hooks to symlink. Either to pollute the PATH namespace with traditional PsN-styled
"binscripts", as::

    /usr/bin/execute               ->  /usr/bin/pharmpy execute

Or, to keep multiple versions when upgrading Pharmpy.

On Logging
----------

As a CLI we are at the top, interacting with the user. All of Pharmpy will get messages
filtered through here (loggers organize themselves in a hiearchy based on their period-split names).

.. warning::
    Loggers **mustn't** be configurated if imported as library (e.g. set handlers and such),
    since that should bubble up into the lap of whomever is at the top (of the Python
    period-separated namespace).

Definitions
===========
"""

import argparse
import importlib
import logging
import pathlib
import pydoc
import re
import sys
import types
from collections import OrderedDict, namedtuple
from textwrap import dedent


class LazyLoader(types.ModuleType):
    """Class that masquerades as a module and lazily loads it when accessed the first time

       The code for the class is taken from TensorFlow and is under the Apache 2.0 license

       This is needed for the cli to be able to do things as version and help quicker than
       if all modules were imported.
    """
    def __init__(self, local_name, parent_module_globals, name):
        self._local_name = local_name
        self._parent_module_globals = parent_module_globals

        super(LazyLoader, self).__init__(name)

    def _load(self):
        # Import the target module and insert it into the parent's namespace
        module = importlib.import_module(self.__name__)
        self._parent_module_globals[self._local_name] = module

        # Update this object's dict so that if someone keeps a reference to the
        #   LazyLoader, lookups are efficient (__getattr__ is only called on lookups
        #   that fail).
        self.__dict__.update(module.__dict__)
        return module

    def __getattr__(self, item):
        module = self._load()
        return getattr(module, item)

    def __dir__(self):
        module = self._load()
        return dir(module)


pharmpy = LazyLoader('pharmpy', globals(), 'pharmpy')
np = LazyLoader('numpy', globals(), 'numpy')
iterators = LazyLoader('iterators', globals(), 'pharmpy.data.iterators')
plugin_utils = LazyLoader('plugin_utils', globals(), 'pharmpy.plugins.utils')

formatter = argparse.ArgumentDefaultsHelpFormatter


def error(exception):
    """Raise the exception with no traceback printed

    Used for non-recoverables in CLI. Exceptions giving traceback are bugs!
    """
    def exc_only_hook(exception_type, exception, traceback):
        print(f'{exception_type.__name__}: {exception}')

    sys.excepthook = exc_only_hook
    raise exception


def results_bootstrap(args):
    """Subcommand to generate bootstrap results"""
    if len(args.models) < 3:
        error(ValueError('Need at least the original model and 2'
                         'other models'))
    res = pharmpy.results.BootstrapResults(original_model=args.models[0],
                                           bootstrap_models=args.models[1:])
    print(res)


def results_frem(args):
    """Generate frem results"""
    from pharmpy.methods.frem.results import psn_frem_results
    if not args.psn_dir.is_dir():
        error(FileNotFoundError(str(args.psn_dir)))
    res = psn_frem_results(args.psn_dir, method=args.method,
                           force_posdef_covmatrix=args.force_posdef_covmatrix,
                           force_posdef_samples=args.force_posdef_samples)
    res.add_plots()
    res.to_json(path=args.psn_dir / 'results.json')
    res.to_csv(path=args.psn_dir / 'results.csv')


def results_ofv(args):
    """Subcommand to extract final ofv from multiple results"""
    ofvs = []
    for model in args.models:
        try:
            ofv = str(model.modelfit_results.ofv)
        except AttributeError:
            ofv = 'NA'
        ofvs.append(ofv)
    print(' '.join(ofvs))


def results_print(args):
    """Subcommand to print any results"""
    if args.dir.is_dir():
        path = args.dir / 'results.json'
    elif args.dir.is_file():
        path = args.dir
    else:
        error(FileNotFoundError(str(args.dir)))
    from pharmpy.results import read_results
    res = read_results(path)
    print(res)


def results_report(args):
    """Subcommand to generate reports"""
    results_path = args.psn_dir / 'results.json'
    if not results_path.is_file():
        error(FileNotFoundError(results_path))
    from pharmpy.results import read_results
    res = read_results(results_path)
    res.create_report(args.psn_dir)


def results_summary(args):
    """Subcommand to output summary of modelfit"""
    for model in args.models:
        print(model.name)
        model.modelfit_results.reparameterize([
            pharmpy.random_variables.NormalParametrizationSd,
            pharmpy.random_variables.MultivariateNormalParametrizationSdCorr
        ])
        print(model.modelfit_results.parameter_summary())
        print()


def check_input_path(path):
    """Resolves path to input file and checks existence.

    Raises if not found or is dir, without tracebacks (see :func:`error_exit`).
    """
    try:
        path = pathlib.Path(path)
        path = path.resolve()
    except FileNotFoundError:
        pass

    if not path.exists():
        exc = FileNotFoundError('No such input file: %r' % str(path))
        error(exc)
    elif path.is_dir():
        exc = IsADirectoryError('Is a directory (not an input file): %r' % str(path))
        error(exc)
    else:
        return path


def input_model(path):
    """Returns :class:`~pharmpy.model.Model` from *path*.

    Raises if not found or is dir, without tracebacks (see :func:`error_exit`).
    """
    path = check_input_path(path)
    return pharmpy.Model(path)


args_input = argparse.ArgumentParser(add_help=False)
group_input = args_input.add_argument_group(title='inputs')
group_input.add_argument('models', metavar='FILE', type=input_model, nargs='+',
                         help='input model files')

parser_definition = [
    {'results': {'subs': [
        {'bootstrap': {'help': 'Generate bootstrap results', 'func': results_bootstrap,
                       'parents': [args_input]}},
        {'frem': {'help': 'Generate FREM results', 'func': results_frem, 'args': [
            {'name': 'psn_dir', 'metavar': 'PsN directory', 'type': pathlib.Path,
             'help': 'Path to PsN frem run directory'},
            {'name': '--method', 'choices': ['cov_sampling', 'bipp'], 'default': 'cov_sampling',
             'help': 'Method to use for uncertainty of covariate effects'},
            {'name': '--force_posdef_covmatrix', 'action': 'store_true',
             'help': 'Should covariance matrix be forced to become positive definite'},
            {'name': '--force_posdef_samples', 'type': int, 'default': 500,
             'help': 'Number of sampling tries to do before starting to force posdef'}]}},
        {'ofv': {'help': 'Extract OFVs from model runs', 'parents': [args_input],
                 'func': results_ofv}},
        {'print': {'help': 'Print results', 'func': results_print, 'args': [
            {'name': 'dir', 'metavar': 'file or directory', 'type': pathlib.Path,
             'help': 'Path to directory containing results.json '
                     'or directly to json results file'}]}},
        {'report': {'help': 'Generate results report', 'func': results_report, 'args': [
            {'name': 'psn_dir', 'metavar': 'PsN directory', 'type': pathlib.Path,
             'help': 'Path to PsN run directory'}]}},
        {'summary': {'help': 'Modelfit summary', 'parents': [args_input],
                     'func': results_summary}}],
        'help': 'Result extraction and generation',
        'title': 'Pharmpy result generation commands', 'metavar': 'ACTION'}}]


def generate_parsers(parsers):
    for command in parser_definition:
        (cmd_name, cmd_dict), = command.items()
        cmd_parser = parsers.add_parser(cmd_name, allow_abbrev=True, help=cmd_dict['help'],
                                        formatter_class=formatter)
        subs = cmd_parser.add_subparsers(title=cmd_dict['title'], metavar=cmd_dict['metavar'])
        for sub_command in cmd_dict['subs']:
            (sub_name, sub_dict), = sub_command.items()
            args = sub_dict.pop('args', [])
            func = sub_dict.pop('func')

            sub_parser = subs.add_parser(sub_name, **sub_dict, formatter_class=formatter)
            for arg in args:
                name = arg.pop('name')
                sub_parser.add_argument(name, **arg)
            sub_parser.set_defaults(func=func)


class CLI:
    """Main CLI interface, based on subcommands (like git).

    Parses and arguments executes all tasks.

    Args:
        *args: Arguments (overrides all of :attr:`sys.argv`).
        subcommand: If given, skip straight to command (sub)parser.
    """

    def __init__(self, *args, subcommand=None):
        """Initialize and parse command line arguments"""

        # root parser and common arguments
        parser = argparse.ArgumentParser(
            prog='pharmpy',
            description=dedent("""
            Welcome to the command line interface of Pharmpy!

            Functionality is split into various subcommands
                - try --help after a COMMAND
                - all keyword arguments can be abbreviated if unique


            """).strip(),
            epilog=dedent("""
                Examples:
                    # Create 100 bootstrap datasets
                    pharmpy data resample pheno_real.mod --resamples=100 --replace

                    # prettyprint model
                    pharmpy model print pheno_real.mod

                    # version/install information
                    pharmpy info
            """).strip(),
            formatter_class=formatter,
            allow_abbrev=True,
        )
        self._init_common_args(parser)
        parser.add_argument('--version', action='version', version='0.3.0')

        # subcommand parsers
        subparsers = parser.add_subparsers(title='Pharmpy commands', metavar='COMMAND')
        self._init_commands_tools(subparsers)
        self._init_commands_misc(subparsers)
        generate_parsers(subparsers)

        # parse
        if args and sys.argv:
            self.logger.debug('Call args overriding, ignoring sys.argv=%r', sys.argv)
        elif not args:
            args = sys.argv[1:]
        if subcommand:
            self.logger.debug('Skipping to subcommand (%r) CLI', subcommand)
            args = [subcommand] + list(args)
        self.logger.debug('%d args to parse: %r' % (len(args), args))
        args = parser.parse_args(args)

        # output/logging setup
        self.set_loglevel(args.loglevel)
        self.logger.debug('Parsed into %d args: %r' % (len(vars(args)), vars(args)))
        if args.hide_traceback:
            self.hide_traceback()

        # dispatch subcommand
        if 'func' in args:
            args.func(args)
        else:
            parser.print_usage()

    # -- additional initializers ---------------------------------------------------------------

    def _init_common_args(self, parser):
        """Initializes common arguments on *parser* and ``self``.

        Groups common to all parsers (activated here)

            .. list-table::
                :widths: 20 80
                :stub-columns: 1

                - - configuration (*idea only*)
                  - To override current (project) configuration file.
                - - logging & verbosity
                  - Sets loglevel and behaviour (e.g. color, logfile copy, etc.) for everything.

        Superclasses for some commands (*parents* in :func:`argparse.ArgumentParser.add_parser`)

            .. list-table::
                :widths: 20 80
                :stub-columns: 1

                - - :attr:`self._args_input`
                  - Common arguments for reading files as input.
                - - :attr:`self._args_output`
                  - Common arguments for outputting files.
        """

        # common to: all
        verbosity = parser.add_argument_group(title='logging & verbosity')
        verbosity.add_argument('--hide_traceback',
                               action='store_true',
                               help="avoid intimidating exception tracebacks")
        loglevel = verbosity.add_mutually_exclusive_group()
        loglevel.add_argument('-v', '--verbose',
                              action='store_const', dest='loglevel',
                              const=logging.INFO, default=logging.WARNING,
                              help='print additional info')
        loglevel.add_argument('-vv', '--debug',
                              action='store_const', dest='loglevel', const=logging.DEBUG,
                              help='show debug messages')

        # common to: commands taking one model file as input
        args_model_input = argparse.ArgumentParser(add_help=False)
        group_model_input = args_model_input.add_argument_group(title='model')
        group_model_input.add_argument('model', metavar='FILE', type=self.input_model,
                                       help='input model file')
        self._args_model_input = args_model_input

        # common to: commands with model file input
        args_input = argparse.ArgumentParser(add_help=False)
        group_input = args_input.add_argument_group(title='inputs')
        group_input.add_argument('models',
                                 metavar='FILE', type=self.input_model, nargs='+',
                                 help='input model files')
        self._args_input = args_input

        # common to: command with model file or dataset input
        args_model_or_data_input = argparse.ArgumentParser(add_help=False)
        group_model_or_data_input = args_model_or_data_input.add_argument_group(title='data_input')
        group_model_or_data_input.add_argument('model_or_dataset',
                                               metavar='FILE', type=self.input_model_or_dataset,
                                               help='input model or dataset file')
        self._args_model_or_data_input = args_model_or_data_input

        # common to: commands with file output
        args_output = argparse.ArgumentParser(add_help=False)
        group_output = args_output.add_argument_group(title='outputs')
        group_output.add_argument('-f', '--force',
                                  action='store_true',
                                  help='remove existing destination files (all)')
        group_output.add_argument('-o', '--output_file',
                                  dest='output_file', metavar='file', type=pathlib.Path,
                                  help='output file')
        self._args_output = args_output

        # common to: commands involving randomization
        args_random = argparse.ArgumentParser(add_help=False)
        group_random = args_random.add_argument_group(title='random seed')
        group_random.add_argument('--seed', type=self.random_seed, metavar='INTEGER',
                                  help='Provide a random seed. The seed must be an integer '
                                  'between 0 and 2^32 - 1')
        self._args_random = args_random

    def _init_commands_tools(self, parsers):
        """Initializes all "tool-like" subcommands."""

        # -- model -----------------------------------------------------------------------------
        cmd_model = parsers.add_parser('model', help='Model manipulations', allow_abbrev=True)

        cmd_model_subs = cmd_model.add_subparsers(title='Pharmpy model commands', metavar='ACTION')

        cmd_model_print = cmd_model_subs.add_parser('print', prog='print model',
                                                    parents=[self._args_input],
                                                    help='Format and print model overview',
                                                    allow_abbrev=True)
        group_apis = cmd_model_print.add_argument_group(title='APIs to print')
        group_apis.add_argument('--all',
                                action='store_true', default=None, help='This is the default')
        group_apis.add_argument('--data',
                                action='store_true',
                                help='A shortened version of the dataset')
        group_apis.add_argument('--parameters',
                                action='store_true',
                                help='Parameters (e.g. inits)')
        group_apis.add_argument('--random_variables',
                                action='store_true',
                                help='Random variables')
        cmd_model_print.set_defaults(func=self.cmd_print)

        cmd_model_sample = cmd_model_subs.add_parser('sample', prog='Sampling of parameter inits',
                                                     parents=[self._args_model_input,
                                                              self._args_random],
                                                     help='Sampling of parameter inits using '
                                                          'uncertainty given by covariance matrix',
                                                     allow_abbrev=True)
        cmd_model_sample.add_argument('--samples', metavar='NUMBER', type=int, default=1,
                                      help='Number of sampled models')
        cmd_model_sample.set_defaults(func=self.model_sample)

        # -- run -------------------------------------------------------------------------------
        cmd_run = parsers.add_parser('run', help='Run workflows', allow_abbrev=True)
        cmd_run_subs = cmd_run.add_subparsers(title='Pharmpy workflow commands', metavar='ACTION')

        cmd_run_execute = cmd_run_subs.add_parser('execute', prog='execute models',
                                                  parents=[self._args_input],
                                                  help='Execute model',
                                                  allow_abbrev=True)
        cmd_run_execute.set_defaults(func=self.run_execute)

        cmd_run_bootstrap = cmd_run_subs.add_parser('bootstrap', prog='bootstrap',
                                                    parents=[self._args_model_input],
                                                    help='Bootstrap',
                                                    allow_abbrev=True)
        cmd_run_bootstrap.set_defaults(func=self.run_bootstrap)

        # -- data ------------------------------------------------------------------------------
        cmd_data = parsers.add_parser('data', help='Data manipulations', allow_abbrev=True)

        cmd_data_subs = cmd_data.add_subparsers(title='Pharmpy data commands', metavar='ACTION')

        cmd_data_write = cmd_data_subs.add_parser('write', help='Write dataset', allow_abbrev=True,
                                                  parents=[self._args_model_or_data_input,
                                                           self._args_output])
        cmd_data_write.set_defaults(func=self.data_write)

        cmd_data_print = cmd_data_subs.add_parser('print', help='Print dataset to stdout',
                                                  allow_abbrev=True,
                                                  parents=[self._args_model_or_data_input])
        cmd_data_print.add_argument('--columns', metavar='COLUMNS', type=self.comma_list,
                                    help='Select specific columns (default is all)')
        cmd_data_print.set_defaults(func=self.data_print)

        cmd_data_filter = cmd_data_subs.add_parser('filter', help='Filter rows of dataset',
                                                   allow_abbrev=True,
                                                   parents=[self._args_model_or_data_input,
                                                            self._args_output])
        cmd_data_filter.add_argument('expressions', nargs=argparse.REMAINDER)
        cmd_data_filter.set_defaults(func=self.data_filter)

        cmd_data_append = cmd_data_subs.add_parser('append', help='Append column to dataset',
                                                   allow_abbrev=True,
                                                   parents=[self._args_model_or_data_input,
                                                            self._args_output])
        cmd_data_append.add_argument('expression')
        cmd_data_append.set_defaults(func=self.data_append)

        cmd_data_resample = cmd_data_subs.add_parser('resample', help='Resample dataset',
                                                     allow_abbrev=True,
                                                     parents=[self._args_model_or_data_input,
                                                              self._args_random])
        cmd_data_resample.add_argument('--group', metavar='COLUMN', type=str, default='ID',
                                       help='Column to use for grouping (default is ID)')
        cmd_data_resample.add_argument('--resamples', metavar='NUMBER', type=int, default=1,
                                       help='Number of resampled datasets (default 1)')
        cmd_data_resample.add_argument('--stratify', metavar='COLUMN', type=str,
                                       help='Column to use for stratification')
        cmd_data_resample.add_argument('--replace', action='store_true',
                                       help='Sample with replacement (default is without)')
        cmd_data_resample.add_argument('--sample_size', metavar='NUMBER', type=int, default=None,
                                       help='Number of groups to sample for each resample')
        cmd_data_resample.set_defaults(func=self.data_resample)

        cmd_data_anon = cmd_data_subs.add_parser('anonymize', help='Anonymize dataset by reorder'
                                                 ' and renumber a specific data column',
                                                 description='Anonymize dataset by resampling '
                                                             'of the group column',
                                                 allow_abbrev=True,
                                                 parents=[self._args_output])
        cmd_data_anon.add_argument('dataset', metavar='FILE', type=str,
                                   help='A csv file dataset')
        cmd_data_anon.add_argument('--group', metavar='COLUMN', type=str, default='ID',
                                   help='column to anonymize (default ID)')
        cmd_data_anon.set_defaults(func=self.data_anonymize)

    def _init_commands_misc(self, parsers):
        """Initializes miscellanelous other (non-tool) subcommands."""

        # -- info ---------------------------------------------------------------------------
        cmd_info = parsers.add_parser('info', prog='pharmpy info',
                                      help='Show pharmpy information',
                                      allow_abbrev=True)
        cmd_info.set_defaults(func=self.cmd_info)

    # -- subcommand launchers ------------------------------------------------------------------

    def cmd_print(self, args):
        """Subcommand for formatting/printing model components."""
        if not (args.data or args.parameters) and args.all is None:
            args.all = True

        lines = []
        for i, model in enumerate(args.models):
            lines += ['[%d/%d] %r' % (i+1, len(args.models), model.name)]
            dict_ = OrderedDict()
            if args.data or args.all:
                dict_['dataset'] = repr(model.input.dataset)
            if args.parameters or args.all:
                s = repr(model.parameters) + '\n\n'
                dict_['parameters'] = s
            if args.random_variables or args.all:
                s = repr(model.random_variables) + '\n\n'
                dict_['random_variables'] = s
            dict_lines = self.format_keyval_pairs(dict_, sort=False)
            lines += ['\t%s' % line for line in dict_lines]
        if len(lines) > 24:
            pydoc.pager('\n'.join(lines))
        else:
            print('\n'.join(lines))

    def model_sample(self, args):
        model = args.model
        from pharmpy.parameter_sampling import sample_from_covariance_matrix
        samples = sample_from_covariance_matrix(model, n=args.samples)
        for row, params in samples.iterrows():
            model.parameters = params
            model.name = f'sample_{row + 1}'
            model.write()

    def run_execute(self, args):
        """A regular execute of one or more models"""
        from pharmpy.workflows.modelfit import ModelfitWorkflow
        wf = ModelfitWorkflow(args.models)
        wf.start()

    def run_bootstrap(self, args):
        from pharmpy.workflows.bootstrap import BootstrapWorkflow
        wf = BootstrapWorkflow(args.model)
        wf.start()

    def data_write(self, args):
        """Subcommand to write a dataset."""
        try:
            df = args.model_or_dataset.input.dataset
        except plugin_utils.PluginError:
            df = args.model_or_dataset
        path = args.output_file
        try:
            # If no output_file supplied will use name of df
            path = df.pharmpy.write_csv(path=path, force=args.force)
            print(f'Dataset written to {path}')
        except OSError as e:
            self.error_exit(exception=e)

    def data_print(self, args):
        """Subcommand to print a dataset to stdout via pager
        """
        try:
            df = args.model_or_dataset.input.dataset
        except AttributeError:
            df = args.model_or_dataset
        if args.columns:
            try:
                df = df[args.columns]
            except KeyError as e:
                self.error_exit(exception=e)
        pydoc.pager(df.to_string())

    def data_filter(self, args):
        """Subcommand to filter a dataset"""
        try:
            df = args.model_or_dataset.input.dataset
        except plugin_utils.PluginError:
            df = args.model_or_dataset
        expression = ' '.join(args.expressions)
        try:
            df.query(expression, inplace=True)
        except SyntaxError:
            self.error_exit(SyntaxError(f'Invalid syntax of query: "{expression}"'))
        except BaseException as e:
            self.error_exit(exception=e)
        self.write_model_or_dataset(args.model_or_dataset, df, args.output_file, args.force)

    def data_append(self, args):
        """Subcommand to append a column to a dataset"""
        try:
            df = args.model_or_dataset.input.dataset
        except plugin_utils.PluginError:
            df = args.model_or_dataset
        try:
            df.eval(args.expression, inplace=True)
        except SyntaxError:
            self.error_exit(SyntaxError(f'Invalid syntax of expression: "{args.expression}"'))
        self.write_model_or_dataset(args.model_or_dataset, df, args.output_file, args.force)

    def write_model_or_dataset(self, model_or_dataset, new_df, path, force):
        """Write model or dataset to output_path or using default names
        """
        if hasattr(model_or_dataset, 'pharmpy'):
            # Is a dataset
            try:
                # If no output_file supplied will use name of df
                path = new_df.pharmpy.write_csv(path=path, force=force)
                print(f'Dataset written to {path}')
            except FileExistsError as e:
                self.error_exit(exception=FileExistsError(f'{e.args[0]} Use -f  or --force to '
                                                          'force an overwrite'))
        else:
            # Is a model
            model = model_or_dataset
            model.input.dataset = new_df
            try:
                if path:
                    if not path.is_dir():
                        self.bump_model_number(model, path)
                    model.write(path=path, force=force)
                else:
                    self.bump_model_number(model)
                    model.write(force=force)
            except FileExistsError as e:
                self.error_exit(exception=FileExistsError(f'{e.args[0]} Use -f or --force to '
                                                          'force an overwrite'))

    def bump_model_number(self, model, path='.'):
        """ If model name ends in a number increase to next available file
            else do nothing.
            FIXME: Could be moved to model class. Need the name of the model and the source object
        """
        path = pathlib.Path(path)
        name = model.name
        m = re.search(r'(.*?)(\d+)$', name)
        if m:
            stem = m.group(1)
            n = int(m.group(2))
            while True:
                n += 1
                new_name = f'{stem}{n}'
                new_path = path / new_name
                if not new_path.exists():
                    break
            model.name = new_name

    def data_resample(self, args):
        """Subcommand to resample a dataset."""
        resampler = iterators.Resample(
                args.model_or_dataset, args.group,
                resamples=args.resamples, stratify=args.stratify, replace=args.replace,
                sample_size=args.sample_size,
                name_pattern=f'{args.model_or_dataset.name}_resample_{{}}')
        for resampled_obj, _ in resampler:
            try:
                try:
                    resampled_obj.write()
                except AttributeError:
                    resampled_obj.pharmpy.write_csv()
            except OSError as e:
                self.error_exit(exception=e)

    def data_anonymize(self, args):
        """Subcommand to anonymize a dataset
           anonymization is really a special resample case where the specified group
           (usually the ID) is given new random unique values and reordering.
           Will be default try to overwrite original file (but won't unless force)
        """
        path = self.check_input_path(args.dataset)
        dataset = pharmpy.data.read_csv(path, raw=True, parse_columns=[args.group])
        resampler = pharmpy.data.iterators.Resample(dataset, args.group)
        df, _ = next(resampler)
        if args.output_file:
            output_file = args.output_file
        else:
            output_file = path
        self.write_model_or_dataset(df, df, output_file, args.force)

    def cmd_info(self, args):
        """Subcommand to print Pharmpy info (and brag a little bit)."""

        lines = self.format_keyval_pairs(self.install._asdict(), right_just=True)
        print('\n'.join(lines))

    # -- helpers -------------------------------------------------------------------------------

    @property
    def logger(self):
        """Returns current CLI logger.

        Initializes, too, if not already configured (= no handlers).
        """

        logger = logging.getLogger(__file__)
        if logger.hasHandlers():
            return logger

        msg = logging.StreamHandler(sys.stdout)
        err = logging.StreamHandler(sys.stderr)

        msg.addFilter(lambda record: record.levelno <= logging.INFO)
        err.setLevel(logging.WARNING)

        logger.addHandler(msg)
        logger.addHandler(err)
        self.set_loglevel(logging.NOTSET)

        return logger

    @property
    def install(self):
        """:class:`namedtuple` with information about install."""
        Install = namedtuple('VersionInfo', ['version', 'authors', 'directory'])
        return Install(pharmpy.__version__,
                       'A list of authors can be found in the AUTHORS.rst',
                       str(pathlib.Path(pharmpy.__file__).parent))

    def random_seed(self, seed):
        try:
            seed = int(seed)
            np.random.seed(seed)
        except ValueError as e:
            # Cannot reraise the ValueError as it is caught by argparse
            self.error_exit(exception=argparse.ArgumentTypeError(str(e)))
        return seed

    def comma_list(self, s):
        if s == '':
            return []
        return s.split(',')

    def input_model(self, path):
        """Returns :class:`~pharmpy.model.Model` from *path*.

        Raises if not found or is dir, without tracebacks (see :func:`error_exit`).
        """
        path = self.check_input_path(path)
        return pharmpy.Model(path)

    def input_model_or_dataset(self, path):
        """Returns :class:`~pharmpy.model.Model` or :class:`~pharmpy.data.PharmDataFrame` from *path*
        """
        path = self.check_input_path(path)
        try:
            obj = pharmpy.Model(path)
        except pharmpy.plugins.utils.PluginError:
            obj = pharmpy.data.read_csv(path)
        return obj

    def check_input_path(self, path):
        """Resolves path to input file and checks existence.

        Raises if not found or is dir, without tracebacks (see :func:`error_exit`).
        """

        try:
            path = pathlib.Path(path)
            path = path.resolve()
        except FileNotFoundError:
            pass

        if not path.exists():
            exc = FileNotFoundError('No such input file: %r' % str(path))
            self.error_exit(exception=exc)
        elif path.is_dir():
            exc = IsADirectoryError('Is a directory (not an input file): %r' % str(path))
            self.error_exit(exception=exc)
        else:
            return path

    def check_output_path(self, path, force):
        """Resolves *path* to output file and checks existence.

        Raises if *path* exists (and not *force*), without tracebacks (see :func:`error_exit`).
        """

        try:
            path = pathlib.Path(path)
            path = path.resolve()
        except FileNotFoundError:
            pass

        if path.exists() and not force:
            msg = "Output file exists: %r\nRefusing to overwrite without --force." % str(path)
            self.error_exit(exception=FileExistsError(msg))
        else:
            return path

    def error_exit(self, msg=None, exception=None, status=1):
        """Error exits with code *status*, with optional *msg* logged and *exception* raised.

        Used for non-recoverables in CLI (raises without tracebacks) irrelevant to pharmpy.
        """

        if msg:
            self.logger.critical(str(msg))
        if exception:
            self.hide_traceback(hide=True)
            raise exception
        else:
            sys.exit(status)

    def format_keyval_pairs(self, data_dict, sort=True, right_just=False):
        """Formats lines from *data_dict*."""

        if not data_dict:
            return []

        key_width = max(len(field) for field in data_dict.keys())
        if sort:
            data_dict = OrderedDict(sorted(data_dict.items()))
        if right_just:
            line_format = '  %%%ds\t%%s' % key_width
        else:
            line_format = '%%-%ds\t%%s' % key_width

        lines = []
        for key, values in data_dict.items():
            if isinstance(values, str):
                values = values.splitlines()
            keys = [key] + ['']*(len(values) - 1)
            for k, v in zip(keys, values):
                lines += [line_format % (k, v)]

        return lines

    # -- configuration -------------------------------------------------------------------------

    def set_loglevel(self, level):
        """Sets current loglevel (and message formatting)."""

        self.logger.setLevel(level)
        if level <= logging.DEBUG and level != logging.NOTSET:
            fmt = "%(asctime)s.%(msecs)03d %(filename)-25s %(lineno)4d %(levelname)-8s %(message)s"
            datefmt = "%H:%M:%S"
        else:
            fmt = '%(message)s'
            datefmt = "%Y-%m-%d %H:%M:%S"
        for handler in self.logger.handlers:
            handler.setFormatter(logging.Formatter(fmt, datefmt))
        self.logger.debug('Logging config: level=logging.%s, fmt=%r, datefmt=%r',
                          logging.getLevelName(level), fmt, datefmt)

    def hide_traceback(self, hide=True, original_hook=sys.excepthook):
        """Override default exception hook to hide all tracebacks.

        .. note:: Even without traceback, the exception and message is shown.
        """

        def exc_only_hook(exception_type, exception, traceback):
            self.logger.critical('%s: %s', exception_type.__name__, exception)

        if hide:
            self.logger.debug('Overriding sys.excepthook to hide tracebacks: %r', sys.excepthook)
            sys.excepthook = exc_only_hook
        else:
            self.logger.debug('Restoring sys.excepthook to reset tracebacks: %r', original_hook)
            sys.excepthook = original_hook


# -- entry point of main CLI (pharmpy) ---------------------------------------------------------


def main(*args):
    """Entry point of ``pharmpy`` CLI util and ``python3 -m pharmpy`` (via ``__main__.py``)."""
    CLI(*args)



# -*- encoding: utf-8 -*-
"""Factory for creating a pharmpy model from a "native" object representation.

Definitions
===========
"""
import pharmpy.source as source
from pharmpy.plugins.utils import detect_model


def Model(obj, **kwargs):
    """Factory for creating a Model object from an object representing the model (i.e. path).

    Arguments:
        obj: Currently a `path-like object`_ pointing to the model file.

    Generic :class:`~pharmpy.generic.Model` if path is None, otherwise appropriate implementation is
    invoked (e.g. NONMEM7 :class:`~pharmpy.api_nonmem.model.Model`).

    .. _path-like object: https://docs.python.org/3/glossary.html#term-path-like-object
    """
    if obj:
        src = source.Source(obj)
        model_class = detect_model(src)
        model = model_class(src, **kwargs)
        return model



"""Pharmpy configuration
"""

# import configparser


class ConfigItem:
    def __init__(self, default, description):
        self.default = default
        self.__doc__ = description

    def __set_name__(self, owner, name):
        self.name = name

    def __get__(self, instance, owner=None):
        return instance.__dict__.get(self.name, self.default)

    def __set__(self, instance, value):
        if type(self.default) != type(value):
            raise TypeError(f'Trying to set configuration item {self.name} using object of wrong '
                            f'type: {type(value)} is not {type(self.default)}')
        instance.__dict__[self.name] = value


class Configuration:
    pass



import pytest


@pytest.mark.usefixtures('parser')
@pytest.mark.parametrize('buf,results', [
    ('$THETA 0', [('THETA(1)', 0, -1000000, 1000000, False)]),
    ('$THETA    12.3 \n\n', [('THETA(1)', 12.3, -1000000, 1000000, False)]),
    ('$THETA  (0,0.00469) ; CL', [('THETA(1)', 0.00469, 0, 1000000, False)]),
    ('$THETA  (0,3) 2 FIXED (0,.6,1) 10 (-INF,-2.7,0)  (37 FIXED)\n 19 (0,1,2)x3', [
        ('THETA(1)', 3, 0, 1000000, False),
        ('THETA(2)', 2, -1000000, 1000000, True),
        ('THETA(3)', .6, 0, 1, False),
        ('THETA(4)', 10, -1000000, 1000000, False),
        ('THETA(5)', -2.7, -1000000, 0, False),
        ('THETA(6)', 37, -1000000, 1000000, True),
        ('THETA(7)', 19, -1000000, 1000000, False),
        ('THETA(8)', 1, 0, 2, False),
        ('THETA(9)', 1, 0, 2, False),
        ('THETA(10)', 1, 0, 2, False),
        ]),
    ('$THETA (0,0.00469555) FIX ; CL', [('THETA(1)', 0.00469555, 0, 1000000, True)]),
    ('$THETA\n ;; Model characteristics\n (0, 0.15, 0.6) ; Proportional\n',
        [('THETA(1)', 0.15, 0, 0.6, False)]),
    ('$THETA\n;\n1;COMMENTING', [('THETA(1)', 1, -1000000, 1000000, False)]),
    ('$THETA\n   ;CMT1;SAMELINE\n;ONLYCOMMENT\n\t;COMMENT2\n    1    \n',
        [('THETA(1)', 1, -1000000, 1000000, False)]),
    ('$THETA\n ;; Model characteristics\n  (0, 0.15, 0.6) ; Proportional error (Drug123)\n'
     '  (0, 1, 10)     ; Additive error (Drug123)\n',
        [('THETA(1)', 0.15, 0, 0.6, False), ('THETA(2)', 1, 0, 10, False)]),
])
def test_parameters(parser, buf, results):
    recs = parser.parse(buf)
    rec = recs.records[0]
    pset = rec.parameters(1)
    assert len(pset) == len(results)
    assert len(pset) == len(rec)
    for res in results:
        name = res[0]
        init = res[1]
        lower = res[2]
        upper = res[3]
        fix = res[4]
        param = pset[name]
        assert param.name == name
        assert param.init == init
        assert param.lower == lower
        assert param.upper == upper
        assert param.fix == fix


def test_theta_num(parser):
    rec = parser.parse('$THETA 1').records[0]
    pset = rec.parameters(2)
    assert len(pset) == 1
    assert pset['THETA(2)'].init == 1


def test_update(parser):
    rec = parser.parse('$THETA 1').records[0]
    pset = rec.parameters(1)
    pset['THETA(1)'].init = 41
    rec.update(pset, 1)
    assert str(rec) == '$THETA 41'

    rec = parser.parse('$THETA 1 FIX').records[0]
    pset = rec.parameters(1)
    pset['THETA(1)'].fix = False
    rec.update(pset, 1)
    assert str(rec) == '$THETA 1'

    rec = parser.parse('$THETA (0.1, 2, 3)  FIX').records[0]
    pset = rec.parameters(1)
    pset['THETA(1)'].fix = False
    rec.update(pset, 1)
    assert str(rec) == '$THETA (0.1, 2, 3)'

    rec = parser.parse('$THETA (0.1, 2, 3 FIX)').records[0]
    pset = rec.parameters(1)
    pset['THETA(1)'].fix = False
    rec.update(pset, 1)
    assert str(rec) == '$THETA (0.1, 2, 3)'

    rec = parser.parse('$THETA (0.1, 2, 3)').records[0]
    pset = rec.parameters(1)
    pset['THETA(1)'].fix = True
    rec.update(pset, 1)
    assert str(rec) == '$THETA (0.1, 2, 3) FIX'

    rec = parser.parse('$THETA 1 2 3 ;CMT').records[0]
    pset = rec.parameters(1)
    pset['THETA(1)'].fix = True
    pset['THETA(3)'].fix = True
    rec.update(pset, 1)
    assert str(rec) == '$THETA 1 FIX 2 3 FIX ;CMT'



from pathlib import Path

import numpy as np
import pandas as pd

from pharmpy.plugins.nonmem.table import NONMEMTableFile
from pharmpy.results import ChainedModelfitResults, ModelfitResults


class NONMEMModelfitResults(ModelfitResults):
    # objects here need to know what chain they are in to be able to do lazy reading
    def __init__(self, chain):
        self._repara = False
        self._chain = chain

    def reparameterize(self, parameterizations):
        names = {p.name for p in parameterizations}
        if len(names) == 2 and names == {'sdcorr', 'sd'}:
            self._repara = True
        else:
            raise NotImplementedError("Only support reparametrization to sdcorr and sd")

    @property
    def ofv(self):
        try:
            return self._ofv
        except AttributeError:
            self._chain._read_ext_table()
            return self._ofv

    @property
    def parameter_estimates(self):
        try:
            if not self._repara:
                return self._parameter_estimates
            else:
                return self._parameter_estimates_sdcorr
        except AttributeError:
            self._chain._read_ext_table()
            if not self._repara:
                return self._parameter_estimates
            else:
                return self._parameter_estimates_sdcorr

    @property
    def covariance_matrix(self):
        """The covariance matrix of the population parameter estimates
        """
        try:
            return self._covariance_matrix
        except AttributeError:
            try:
                self._chain._read_cov_table()
            except OSError:
                pass
            else:
                return self._covariance_matrix
            try:
                self._chain._read_cor_table()
            except OSError:
                pass
            else:
                self._covariance_matrix = self._cov_from_corrse()
                return self._covariance_matrix
            try:
                self._chain._read_coi_table()
            except OSError:
                pass
            else:
                self._covariance_matrix = self._cov_from_inf()
                return self._covariance_matrix
            raise FileNotFoundError("Could not find any of the cov/cor/coi files")

    @property
    def information_matrix(self):
        """The Fischer information matrix of the population parameter estimates
        """
        try:
            return self._information_matrix
        except AttributeError:
            try:
                self._chain._read_coi_table()
            except OSError:
                pass
            else:
                return self._information_matrix
            try:
                self._chain._read_cor_table()
            except OSError:
                pass
            else:
                self._information_matrix = self._inf_from_cov()
                return self._information_matrix
            try:
                self._chain._read_cor_table()
            except OSError:
                pass
            else:
                self._information_matrix = self._inf_from_corrse()
            raise FileNotFoundError("Could not find any of the cov/cor/coi files")

    @property
    def correlation_matrix(self):
        try:
            return self._correlation_matrix
        except AttributeError:
            try:
                self._chain._read_cor_table()
            except OSError:
                pass
            else:
                return self._correlation_matrix
            try:
                self._chain._read_cov_table()
            except OSError:
                pass
            else:
                self._correlation_matrix = self._corr_from_cov()
                return self._correlation_matrix
            try:
                self._chain._read_coi_table()
            except OSError:
                pass
            else:
                self._correlation_matrix = self._corr_from_coi()
                return self._correlation_matrix
            raise FileNotFoundError("Could not find any of the cov/cor/coi files")

    @property
    def standard_errors(self):
        try:
            if not self._repara:
                return self._standard_errors
            else:
                return self._standard_errors_sdcorr
        except AttributeError:
            try:
                self._chain._read_ext_table()
            except OSError:
                pass
            else:
                if not self._repara:
                    return self._standard_errors
                else:
                    return self._standard_errors_sdcorr
            if self._repara:
                raise NotImplementedError("Cannot find ext-file to get standard errors on sdcorr "
                                          "form. Not supported to try getting SEs from other "
                                          "NONMEM output files")
            try:
                self._chain._read_cor_table()
            except OSError:
                pass
            else:
                return self._standard_errors
            try:
                self._chain._read_cov_table()
            except OSError:
                pass
            else:
                self._standard_errors = self._se_from_cov()
                return self._standard_errors
            try:
                self._chain._read_coi_table()
            except OSError:
                pass
            else:
                self._standard_errors = self._se_from_inf()
                return self._standard_errors
            raise FileNotFoundError("Could not find any of the ext/cov/cor/coi files")

    @property
    def individual_ofv(self):
        """A Series with individual estimates indexed over ID
        """
        try:
            return self._individual_ofv
        except AttributeError:
            self._chain._read_phi_table()
            return self._individual_ofv

    @property
    def individual_estimates(self):
        """ETA values from phi-file
        """
        try:
            return self._individual_estimates
        except AttributeError:
            self._chain._read_phi_table()
            return self._individual_estimates

    @property
    def individual_estimates_covariance(self):
        """ETCs from phi-file as Series of DataFrames
        """
        try:
            return self._individual_estimates_covariance
        except AttributeError:
            self._chain._read_phi_table()
            return self._individual_estimates_covariance


class NONMEMChainedModelfitResults(ChainedModelfitResults):
    def __init__(self, path, n, model=None):
        # Path is path to any result file. lst-file would make most sense probably
        # FIXME: n (number of $EST) could be removed as it could be inferred
        self._path = Path(path)
        self._read_phi = False
        self._read_ext = False
        self._read_cov = False
        self._read_coi = False
        self._read_cor = False
        self.model = model
        for i in range(n):
            res = NONMEMModelfitResults(self)
            res.model_name = self._path.stem
            res.model = model
            if i != n - 1:    # Only the final estimation step carries covariance information
                res._covariance_matrix = None
                res._correlation_matrix = None
                res._information_matrix = None
            self.append(res)
        extensions = ['.lst', '.ext', '.cov', '.cor', '.coi', '.phi']
        self.tool_files = [self._path.with_suffix(ext) for ext in extensions]

    def _read_ext_table(self):
        if not self._read_ext:
            ext_tables = NONMEMTableFile(self._path.with_suffix('.ext'))
            for table, result_obj in zip(ext_tables, self):
                ests = table.final_parameter_estimates
                fix = table.fixed
                ests = ests[~fix]
                result_obj._parameter_estimates = ests
                sdcorr = table.omega_sigma_stdcorr[~fix]
                sdcorr_ests = ests.copy()
                sdcorr_ests.update(sdcorr)
                result_obj._parameter_estimates_sdcorr = sdcorr_ests
                try:
                    ses = table.standard_errors
                except Exception:
                    result_obj._standard_errors = None
                else:
                    ses = ses[~fix]
                    result_obj._standard_errors = ses
                    sdcorr = table.omega_sigma_se_stdcorr[~fix]
                    sdcorr_ses = ses.copy()
                    sdcorr_ses.update(sdcorr)
                    result_obj._standard_errors_sdcorr = sdcorr_ses
                result_obj._ofv = table.final_ofv
            self._read_ext = True

    def _read_cov_table(self):
        if not self._read_cov:
            cov_table = NONMEMTableFile(self._path.with_suffix('.cov'))
            self[-1]._covariance_matrix = next(cov_table).data_frame
            self._read_cov = True

    def _read_coi_table(self):
        if not self._read_coi:
            coi_table = NONMEMTableFile(self._path.with_suffix('.coi'))
            self[-1]._information_matrix = next(coi_table).data_frame
            self._read_coi = True

    def _read_cor_table(self):
        if not self._read_cor:
            cor_table = NONMEMTableFile(self._path.with_suffix('.cor'))
            cor = next(cor_table).data_frame
            if not hasattr(self[-1], '_standard_errors'):   # In case read from the ext-file
                self[-1]._standard_errors = pd.Series(np.diag(cor), index=cor.index)
            np.fill_diagonal(cor.values, 1)
            self[-1]._correlation_matrix = cor
            self._read_cor = True

    def _read_phi_table(self):
        if not self._read_phi:
            rv_names = [rv.name for rv in self.model.random_variables if rv.name.startswith('ETA')]
            phi_tables = NONMEMTableFile(self._path.with_suffix('.phi'))
            for table, result_obj in zip(phi_tables, self):
                result_obj._individual_ofv = table.iofv
                result_obj._individual_estimates = table.etas[rv_names]
                covs = table.etcs
                for index, cov in covs.iteritems():     # Remove ETCs for 0 FIX OMEGAs
                    covs[index] = cov[rv_names].loc[rv_names]
                result_obj._individual_estimates_covariance = covs

            self._read_phi = True



"""
===============
Model Execution
===============

Central support for executing models for evaluation, estimation and simulation tasks. Useful for:

    - :class:`~pharmpy.model.Model` objects to, for example, evaluate themselves on request
    - Full workflows in implementations of :class:`~pharmpy.tool.Tool` (e.g. FREM)

This package defines four classes:

    .. list-table::
        :widths: 20 80
        :stub-columns: 1

        - - :class:`~pharmpy.execute.job.Job`
          - A job unit.
            Contains an asynchronous process and exposes blocking and non-blocking stdout.
        - - :class:`~pharmpy.execute.run_directory.RunDirectory`
          - Run directory with cleaning and safe unlink operations.
            Invoking directory, where are the models? Which files to copy where?
        - - :class:`~pharmpy.execute.environment.Environment`
          - Platform (e.g. Linux) & system (e.g. SLURM) implementation.
            The cluster/local or OS etc. to start jobs on.
        - - :class:`~pharmpy.execute.engine.Engine`
          - Creates Job objects, executing some task, for a *specific* implementation.
            Contains Environment object. The focal point for implementation inheritance.

Where :class:`~pharmpy.execute.engine.Engine` is the **critical unit** for a implementation to
inherit, e.g. :class:`~pharmpy.api_nonmem.execute.NONMEM7`.

:class:`~pharmpy.execute.environment.Environment` require inheritance for a specific purpose (e.g.
Linux, SLURM, dummy directory, etc.), but not an implementation.
:class:`~pharmpy.execute.run_directory.RunDirectory` and :class:`~pharmpy.execute.job.Job` will
likely remain model (type) agnostic.

Definitions
-----------
"""

from .engine import Engine
from .environment import Environment

__all__ = ['Engine', 'Environment']



from collections import OrderedDict


def test_create_record(parser):
    recs = parser.parse('$INPUT ID TIME DV WGT=DROP')
    rec = recs.records[0]
    pairs = rec.option_pairs
    assert pairs == OrderedDict([('ID', None), ('TIME', None), ('DV', None), ('WGT', 'DROP')])

    recs = parser.parse('$EST MAXEVAL=9999 INTERACTION')
    rec = recs.records[0]
    pairs = rec.option_pairs
    assert pairs == OrderedDict([('MAXEVAL', '9999'), ('INTERACTION', None)])


def test_all_options(parser):
    recs = parser.parse('$INPUT ID ID TIME DV WGT=DROP')
    rec = recs.records[0]
    pairs = rec.all_options
    assert pairs == [('ID', None), ('ID', None), ('TIME', None), ('DV', None), ('WGT', 'DROP')]


def test_set_option(parser):
    rec = parser.parse('$ETAS FILE=run1.phi').records[0]
    rec.set_option("FILE", "new.phi")
    assert rec.option_pairs == OrderedDict([('FILE', 'new.phi')])
    assert str(rec) == '$ETAS FILE=new.phi'

    rec = parser.parse('$EST METHOD=1 INTER ; my est\n').records[0]
    rec.set_option("METHOD", "0")
    assert str(rec) == '$EST METHOD=0 INTER ; my est\n'
    rec.set_option("CTYPE", "4")
    assert str(rec) == '$EST METHOD=0 INTER CTYPE=4 ; my est\n'



import pandas as pd
import pytest

from pharmpy.plugins.nonmem.results import NONMEMChainedModelfitResults


def test_ofv(pheno_lst):
    res = NONMEMChainedModelfitResults(pheno_lst, 1)
    assert res.ofv == 586.27605628188053


def test_tool_files(pheno_lst):
    res = NONMEMChainedModelfitResults(pheno_lst, 1)
    names = [str(p.name) for p in res.tool_files]
    assert names == ['pheno_real.lst', 'pheno_real.ext', 'pheno_real.cov', 'pheno_real.cor',
                     'pheno_real.coi', 'pheno_real.phi']


def test_covariance(pheno_lst):
    res = NONMEMChainedModelfitResults(pheno_lst, 1)
    cov = res.covariance_matrix
    assert len(cov) == 6
    assert pytest.approx(cov.loc['THETA(1)', 'THETA(1)'], 1e-13) == 4.41151E-08
    assert pytest.approx(cov.loc['OMEGA(2,2)', 'THETA(2)'], 1e-13) == 7.17184E-05


def test_information(pheno_lst):
    res = NONMEMChainedModelfitResults(pheno_lst, 1)
    m = res.information_matrix
    assert len(m) == 6
    assert pytest.approx(m.loc['THETA(1)', 'THETA(1)'], 1e-13) == 2.99556E+07
    assert pytest.approx(m.loc['OMEGA(2,2)', 'THETA(2)'], 1e-13) == -2.80082E+03


def test_correlation(pheno_lst):
    res = NONMEMChainedModelfitResults(pheno_lst, 1)
    corr = res.correlation_matrix
    assert len(corr) == 6
    assert corr.loc['THETA(1)', 'THETA(1)'] == 1.0
    assert pytest.approx(corr.loc['OMEGA(2,2)', 'THETA(2)'], 1e-13) == 3.56662E-01


def test_standard_errors(pheno_lst):
    res = NONMEMChainedModelfitResults(pheno_lst, 1)
    ses = res.standard_errors
    assert len(ses) == 6
    assert pytest.approx(ses['THETA(1)'], 1e-13) == 2.10036E-04


def test_individual_ofv(pheno, pheno_lst):
    res = NONMEMChainedModelfitResults(pheno_lst, 1, model=pheno)
    iofv = res.individual_ofv
    assert len(iofv) == 59
    assert pytest.approx(iofv[1], 1e-15) == 5.9473520242962552
    assert pytest.approx(iofv[57], 1e-15) == 5.6639479151436394
    assert res.plot_iofv_vs_iofv(res)


def test_individual_estimates(pheno, pheno_lst):
    res = NONMEMChainedModelfitResults(pheno_lst, 1, model=pheno)
    ie = res.individual_estimates
    assert len(ie) == 59
    assert pytest.approx(ie['ETA(1)'][1], 1e-15) == -0.0438608
    assert pytest.approx(ie['ETA(2)'][1], 1e-15) == 0.00543031
    assert pytest.approx(ie['ETA(1)'][28], 1e-15) == 7.75957e-04
    assert pytest.approx(ie['ETA(2)'][28], 1e-15) == 8.32311E-02


def test_individual_shrinkage(pheno, pheno_lst):
    res = NONMEMChainedModelfitResults(pheno_lst, 1, model=pheno)
    ishr = res.individual_shrinkage
    assert len(ishr) == 59
    assert pytest.approx(ishr['ETA(1)'][1], 1e-15) == 0.84778949807160287


def test_eta_shrinkage(pheno, pheno_lst):
    res = NONMEMChainedModelfitResults(pheno_lst, 1, model=pheno)
    shrinkage = res.eta_shrinkage()
    assert len(shrinkage) == 2
    assert pytest.approx(shrinkage['ETA(1)'], 0.0001) == 7.2048E+01 / 100
    assert pytest.approx(shrinkage['ETA(2)'], 0.0001) == 2.4030E+01 / 100
    shrinkage = res.eta_shrinkage(sd=True)
    assert len(shrinkage) == 2
    assert pytest.approx(shrinkage['ETA(1)'], 0.0001) == 4.7130E+01 / 100
    assert pytest.approx(shrinkage['ETA(2)'], 0.0001) == 1.2839E+01 / 100


def test_individual_estimates_covariance(pheno, pheno_lst):
    res = NONMEMChainedModelfitResults(pheno_lst, 1, model=pheno)
    cov = res.individual_estimates_covariance
    assert len(cov) == 59
    names = ['ETA(1)', 'ETA(2)']
    correct = pd.DataFrame([[2.48833E-02, -2.99920E-03], [-2.99920E-03, 7.15713E-03]],
                           index=names, columns=names)
    pd.testing.assert_frame_equal(cov[1], correct)
    correct2 = pd.DataFrame([[2.93487E-02, -1.95747E-04], [-1.95747E-04, 8.94118E-03]],
                            index=names, columns=names)
    pd.testing.assert_frame_equal(cov[43], correct2)


def test_parameter_estimates(pheno_lst):
    res = NONMEMChainedModelfitResults(pheno_lst, 1)
    pe = res.parameter_estimates
    assert len(pe) == 6
    assert pe['THETA(1)'] == 4.69555e-3
    assert pe['OMEGA(2,2)'] == 2.7906e-2



import pytest
import sympy
import sympy.stats as stats

from pharmpy.random_variables import JointNormalSeparate, RandomVariables


def test_joint_normal_separate():
    rvs = JointNormalSeparate(['ETA(1)', 'ETA(2)'], [0, 0], [[1, 0], [0, 1]])
    assert rvs[0].name == 'ETA(1)'
    assert rvs[1].name == 'ETA(2)'
    assert rvs[0].pspace.distribution.mu == sympy.Matrix([[0], [0]])
    assert rvs[0].pspace.distribution.sigma == sympy.Matrix([[1, 0], [0, 1]])
    assert stats.random_symbols(rvs[0]) == [rvs[0]]
    assert stats.random_symbols(rvs[1]) == [rvs[1]]

    rvs2 = JointNormalSeparate(['ETA(3)', 'ETA(4)'], [1, 1], [[6, 3], [4, 5]])
    # Check that previously created rvs are still intact
    assert rvs[0].name == 'ETA(1)'
    assert rvs[1].name == 'ETA(2)'
    assert rvs[0].pspace.distribution.mu == sympy.Matrix([[0], [0]])
    assert rvs[0].pspace.distribution.sigma == sympy.Matrix([[1, 0], [0, 1]])
    assert stats.random_symbols(rvs[0]) == [rvs[0]]
    assert stats.random_symbols(rvs[1]) == [rvs[1]]
    assert len(rvs2) == 2


def test_rv():
    omega1 = sympy.Symbol('OMEGA(1,1)')
    x = stats.Normal('ETA(1)', 0, sympy.sqrt(omega1))
    rvs = RandomVariables([x])
    assert len(rvs) == 1
    retrieved = rvs['ETA(1)']
    assert retrieved.name == 'ETA(1)'
    assert retrieved.pspace.distribution.mean == 0


def test_distributions():
    rvs = JointNormalSeparate(['ETA(1)', 'ETA(2)'], [0, 0], [[3, 0.25], [0.25, 1]])
    rvs = RandomVariables(rvs)
    rvs.add(stats.Normal('ETA(3)', 0.5, 2))
    gen = rvs.distributions()
    symbols, dist = next(gen)
    assert symbols[0].name == 'ETA(1)'
    assert symbols[1].name == 'ETA(2)'
    assert len(symbols) == 2
    assert dist == rvs[0].pspace.distribution
    symbols, dist = next(gen)
    assert symbols[0].name == 'ETA(3)'
    assert len(symbols) == 1
    assert dist == rvs[2].pspace.distribution
    with pytest.raises(StopIteration):
        symbols, dist = next(gen)


def test_merge_normal_distributions():
    rvs = JointNormalSeparate(['ETA(1)', 'ETA(2)'], [0, 0], [[3, 0.25], [0.25, 1]])
    rvs = RandomVariables(rvs)
    rvs.add(stats.Normal('ETA(3)', 0.5, 2))
    rvs.merge_normal_distributions()
    assert len(rvs) == 3
    assert rvs['ETA(1)'].name == 'ETA(1)'
    assert rvs[1].name == 'ETA(2)'
    assert rvs[2].name == 'ETA(3)'
    assert rvs[0].pspace is rvs[1].pspace
    assert rvs[0].pspace is rvs[2].pspace
    dist = rvs[0].pspace.distribution
    assert dist.mu == sympy.Matrix([0, 0, 0.5])
    assert dist.sigma == sympy.Matrix([[3, 0.25, 0], [0.25, 1, 0], [0, 0, 4]])
    rvs.merge_normal_distributions(fill=1)
    dist = rvs[0].pspace.distribution
    assert dist.sigma == sympy.Matrix([[3, 0.25, 1], [0.25, 1, 1], [1, 1, 4]])


def test_validate_parameters():
    a, b, c, d = sympy.symbols('a b c d')
    rvs = JointNormalSeparate(['ETA(1)', 'ETA(2)'], [0, 0], [[a, b], [b, c]])
    rvs = RandomVariables(rvs)
    rvs.add(stats.Normal('ETA(3)', 0.5, d))
    params = {'a': 2, 'b': 0.1, 'c': 1, 'd': 23}
    assert rvs.validate_parameters(params)
    params2 = {'a': 2, 'b': 2, 'c': 1, 'd': 23}
    assert not rvs.validate_parameters(params2)



import asyncio
import logging
import sys
import textwrap

import pytest


@pytest.fixture(scope='function')
def py_command():
    code = textwrap.dedent("""
        import sys
        print('OUT', file=sys.stdout, flush=True)
        print('ERR', file=sys.stderr, flush=True)
    """)
    command = [sys.executable, '-c', code]

    output = dict(output=['OUT', 'ERR'])
    output['stdout'] = list(filter(lambda x: x.startswith('OUT'), output))
    output['stderr'] = list(filter(lambda x: x.startswith('ERR'), output))

    return (command, output)


@pytest.fixture(scope='function')
def py_command_slow():
    code = textwrap.dedent("""
        import sys
        import time

        for i in range(3):
            print('OUT[%d]' % (i+1), file=sys.stdout, flush=True)
            print('ERR[%d]' % (i+1), file=sys.stderr, flush=True)
            time.sleep(0.05)
    """)
    command = [sys.executable, '-c', code]

    output = textwrap.dedent("""
    OUT[1]
    ERR[1]
    OUT[2]
    ERR[2]
    OUT[3]
    ERR[3]
    """).strip().splitlines()
    output = dict(output=output)
    output['stdout'] = list(filter(lambda x: x.startswith('OUT'), output['output']))
    output['stderr'] = list(filter(lambda x: x.startswith('ERR'), output['output']))

    return (command, output)


def debuglog_jobs(section_title, jobs, logger=None):
    """Log debugging information on jobs, current event loop and child watcher."""

    if not logger:
        logger = logging.getLogger()
    logger.warning('DEBUG (%s)', section_title.upper())

    event_loop = asyncio.get_event_loop()
    logger.warning('event_loop=%r (%s)', event_loop, hex(id(event_loop)))
    for i, job in enumerate(jobs):
        job.init.wait()
        logger.warning('jobs[%d]: is_alive=%r job=%r', i, job.is_alive(), job)
        logger.warning('jobs[%d]: job.loop=%r (%s)', i, job.loop, hex(id(job.loop)))

    watcher = asyncio.get_child_watcher()
    logger.warning('watcher=%r: loop=%r (%s)', watcher, watcher._loop, hex(id(watcher._loop)))
    for pid, cb in watcher._callbacks.items():
        func = cb[0].__func__
        transport = cb[1][0]
        logger.warning('watcher._callbacks[%i] %r', pid, func)
        logger.warning('watcher._callbacks[%i] %r', pid, transport)



import pytest
import sympy
from sympy import Symbol


def S(x):
    return Symbol(x, real=True)


@pytest.mark.usefixtures('parser')
@pytest.mark.parametrize('buf,symbol,expression', [
    ('$PRED\nY = THETA(1) + ETA(1) + EPS(1)', S('Y'), S('THETA(1)') + S('ETA(1)') + S('EPS(1)')),
    ('$PRED\nCL = 2', S('CL'), 2),
    ('$PRED K=-1', S('K'), -1),
    ('$PRED K=-1.5', S('K'), -1.5),
    ('$PRED\nCL = KA', S('CL'), S('KA')),
    ('$PRED\nG = BASE - LESS', S('G'), S('BASE') - S('LESS')),
    ('$PRED CL = THETA(1) * LEFT', S('CL'), S('THETA(1)') * S('LEFT')),
    ('$PRED D2 = WGT / SRC', S('D2'), S('WGT') / S('SRC')),
    ('$PRED D = W * B + C', S('D'), S('W') * S('B') + S('C')),
    ('$PRED D = W * (B + C)', S('D'), S('W') * (S('B') + S('C'))),
    ('$PRED D = A+B*C+D', S('D'), S('A') + (S('B') * S('C')) + S('D')),
    ('$PRED D = A**2', S('D'), S('A') ** 2),
    ('$PRED D = A - (-2)', S('D'), S('A') + 2),
    ('$PRED D = A - (+2)', S('D'), S('A') - 2),
    ('$PRED D = 2.5', S('D'), 2.5),
    ('$PRED CL = EXP(2)', S('CL'), sympy.exp(2)),
    ('$PRED CL = LOG(V + 1)', S('CL'), sympy.log(S('V') + 1)),
    ('$PRED CL = LOG10(3.5 + THETA(1))', S('CL'), sympy.log(3.5 + S('THETA(1)'), 10)),
    ('$PRED CL = (SIN(X) + COS(X))', S('CL'), sympy.sin(S('X')) + sympy.cos(S('X'))),
    ('$PRED S22 = ABS(1 + 2 + SIN(X))', S('S22'), sympy.Abs(3 + sympy.sin(S('X')))),
    ('$PRED CL = TAN(X) * EXP(Y)', S('CL'), sympy.tan(S('X')) * sympy.exp(S('Y'))),
    ('$PRED K_ = ATAN(1) - ASIN(X)/ACOS(X)', S('K_'), sympy.atan(1) - sympy.asin(S('X')) /
        sympy.acos(S('X'))),
    ('$PRED CL = INT(-2.2)', S('CL'), -2),
    ('$PRED CL = INT(0.2)', S('CL'), 0),
    ('$PRED CL = MOD(1, 2)', S('CL'), sympy.Mod(1, 2)),
    ('$PRED CL = GAMLN(2 + X)   ;COMMENT', S('CL'), sympy.loggamma(S('X') + 2)),
    ('$PRED IF (X.EQ.2) CL=23', S('CL'), sympy.Piecewise((23, sympy.Eq(S('X'), 2)))),
    ('$PRED IF (X.NE.1.5) CL=THETA(1)', S('CL'),
        sympy.Piecewise((S('THETA(1)'), sympy.Ne(S('X'), 1.5)))),
    ('$PRED IF (X.EQ.2+1) CL=23', S('CL'), sympy.Piecewise((23, sympy.Eq(S('X'), 3)))),
    ('$PRED IF (X < ETA(1)) CL=23', S('CL'), sympy.Piecewise((23, sympy.Lt(S('X'), S('ETA(1)'))))),
    ('$PK IF(AMT.GT.0) BTIME=TIME', S('BTIME'),
        sympy.Piecewise((S('TIME'), sympy.Gt(S('AMT'), 0)))),
    ('$PRED IF (X.EQ.2.AND.Y.EQ.3) CL=23', S('CL'),
        sympy.Piecewise((23, sympy.And(sympy.Eq(S('X'), 2), sympy.Eq(S('Y'), 3))))),
    ('$PRED IF (X.EQ.2.OR.Y.EQ.3) CL=23', S('CL'),
        sympy.Piecewise((23, sympy.Or(sympy.Eq(S('X'), 2), sympy.Eq(S('Y'), 3))))),
    ('$PRED IF (.NOT.X.EQ.2) CL=25', S('CL'),
        sympy.Piecewise((25, sympy.Not(sympy.Eq(S('X'), 2))))),
    ('$PRED IF (Q.EQ.(R+C)/D) L=0', S('L'),
        sympy.Piecewise((0, sympy.Eq(S('Q'), sympy.Mul(sympy.Add(S('R'), S('C')), 1 / S('D')))))),
    ('$PRED IF (Q.EQ.R+C/D) L=0', S('L'),
        sympy.Piecewise((0, sympy.Eq(S('Q'), S('R') + S('C') / S('D'))))),
])
def test_single_assignments(parser, buf, symbol, expression):
    rec = parser.parse(buf).records[0]
    assert len(rec.statements) == 1
    assert rec.statements[0].symbol == symbol
    assert rec.statements[0].expression == expression


@pytest.mark.usefixtures('parser')
@pytest.mark.parametrize('buf,symb_expr_arr', [
    ('$PRED\nIF (X.EQ.0) THEN\nY = 23\nZ = 9\nEND IF', [
        (S('Y'), sympy.Piecewise((23, sympy.Eq(S('X'), 0)))),
        (S('Z'), sympy.Piecewise((9, sympy.Eq(S('X'), 0))))]),
    ('$PRED IF (B0.LT.3) THEN\nCL = THETA(1)\nELSE\nCL = 23\nEND IF', [
        (S('CL'), sympy.Piecewise((S('THETA(1)'), S('B0') < 3), (23, True)))]),
    ('$PRED IF (B0.LT.3) THEN\nCL = THETA(1)\nKA = THETA(2)\n  ELSE  \nCL = 23\nKA=82\nEND IF', [
        (S('CL'), sympy.Piecewise((S('THETA(1)'), S('B0') < 3), (23, True))),
        (S('KA'), sympy.Piecewise((S('THETA(2)'), S('B0') < 3), (82, True)))]),
    ('$PRED    IF (A>=0.5) THEN    \n  VAR=1+2 \nELSE IF  (B.EQ.23)  THEN \nVAR=9.25\nEND IF  \n', [
        (S('VAR'), sympy.Piecewise((3, S('A') >= 0.5), (9.25, sympy.Eq(S('B'), 23))))]),
    ('$PRED   IF (A>=0.5) THEN   \n  VAR1=1+2 \nELSE IF  (B.EQ.23)  THEN \nVAR2=9.25\nEND IF  \n', [
        (S('VAR1'), sympy.Piecewise((3, S('A') >= 0.5))),
        (S('VAR2'), sympy.Piecewise((9.25, sympy.Eq(S('B'), 23))))]),
])
def test_block_if(parser, buf, symb_expr_arr):
    rec = parser.parse(buf).records[0]
    assert len(rec.statements) == len(symb_expr_arr)
    for statement, (symb, expr) in zip(rec.statements, symb_expr_arr):
        assert statement.symbol == symb
        assert statement.expression == expr


def test_pheno(parser):
    code = """$PK

IF(AMT.GT.0) BTIME=TIME
TAD=TIME-BTIME
      TVCL=THETA(1)*WGT
      TVV=THETA(2)*WGT
IF(APGR.LT.5) TVV=TVV*(1+THETA(3))
      CL=TVCL*EXP(ETA(1))
      V=TVV*EXP(ETA(2))
      S1=V
"""
    rec = parser.parse(code).records[0]
    assert len(rec.statements) == 8
    assert rec.statements[0].symbol == S('BTIME')
    assert rec.statements[1].symbol == S('TAD')
    assert rec.statements[2].symbol == S('TVCL')
    assert rec.statements[3].symbol == S('TVV')
    assert rec.statements[4].symbol == S('TVV')
    assert rec.statements[5].symbol == S('CL')
    assert rec.statements[6].symbol == S('V')
    assert rec.statements[7].symbol == S('S1')
    assert rec.statements[0].expression == sympy.Piecewise((S('TIME'), sympy.Gt(S('AMT'), 0)))
    assert rec.statements[1].expression == S('TIME') - S('BTIME')
    assert rec.statements[2].expression == S('THETA(1)') * S('WGT')
    assert rec.statements[3].expression == S('THETA(2)') * S('WGT')
    assert rec.statements[4].expression == sympy.Piecewise((S('TVV') * (1 + S('THETA(3)')),
                                                           sympy.Lt(S('APGR'), 5)))
    assert rec.statements[5].expression == S('TVCL') * sympy.exp(S('ETA(1)'))
    assert rec.statements[6].expression == S('TVV') * sympy.exp(S('ETA(2)'))
    assert rec.statements[7].expression == S('V')
    symbol_names = {s.name for s in rec.statements.free_symbols}
    assert symbol_names == {'AMT', 'BTIME', 'TIME', 'TAD', 'TVCL', 'THETA(1)', 'WGT', 'TVV',
                            'THETA(2)', 'APGR', 'THETA(3)', 'CL', 'ETA(1)', 'V', 'ETA(2)', 'S1'}



import enum
from pathlib import Path

import pandas as pd

import pharmpy.data


class DatasetError(Exception):
    pass


class DatasetWarning(Warning):
    pass


class ColumnType(enum.Enum):
    """The type of the data in a column
    """
    UNKNOWN = enum.auto()
    ID = enum.auto()
    IDV = enum.auto()
    DV = enum.auto()
    COVARIATE = enum.auto()

    def __repr__(self):
        return f'{self.__class__.__name__}.{self.name}'

    @property
    def max_one(self):
        """Can this ColumnType only be assigned to at most one column?
        """
        return self in (ColumnType.ID, ColumnType.IDV)


class PharmDataFrame(pd.DataFrame):
    """A DataFrame with additional metadata.

    Each column can have a ColumnType. The default ColumnType is UNKNOWN.

    ============  =============
    ColumnType    Description
    ============  =============
    ID            Individual identifier. Max one per DataFrame. All values have to be unique
    IDV           Independent variable. Max one per DataFrame.
    DV            Dependent variable
    COVARIATE     Covariate
    UNKOWN        Unkown type. This will be the default for columns that hasn't been assigned a type
    ============  =============

    """
    _metadata = ['_column_types', 'name']

    @property
    def _constructor(self):
        return PharmDataFrame

    @property
    def _constructor_sliced(self):
        return pd.Series

    def __deepcopy__(self, memo):
        return self.copy()

    def copy(self, *kwargs):
        """
        """
        # FIXME: Set empty docstring to avoid getting documentation from base class
        #        would like sphinx to do this so that in object docstring is kept.
        new_df = super().copy(*kwargs)
        try:
            new_df._column_types = self._column_types.copy()
        except AttributeError:
            pass
        return new_df


class ColumnTypeIndexer:
    """Indexing a PharmDataFrame to get or set column types
       An instance of ColumnTypeIndexer can be retrieved by df.pharmpy.column_type
    """
    def __init__(self, df):
        self._obj = df

    def _set_one_column_type(self, label, tp):
        if label not in self._obj.columns:
            raise KeyError(str(label))
        if tp.max_one and hasattr(self._obj, '_column_types') and \
                tp in self._obj._column_types.values():
            raise KeyError(f'Only one column of type {tp} is allowed in a PharmDataFrame.')
        try:
            self._obj._column_types[label] = tp
        except AttributeError:
            self._obj._column_types = {}
            self._obj._column_types[label] = tp

    def __setitem__(self, ind, tp):
        if isinstance(ind, str):
            self._set_one_column_type(ind, tp)
        else:
            if hasattr(tp, '__len__') and not len(tp) == 1:
                if len(ind) == len(tp):
                    for label, one_tp in zip(ind, tp):
                        self._set_one_column_type(label, one_tp)
                else:
                    raise ValueError(f'Cannot set {len(ind)} columns using {len(tp)} column types')
            else:
                # Broadcasting of tp
                try:
                    len(ind)
                except TypeError:
                    ind = [ind]
                for label in ind:
                    self._set_one_column_type(label, tp)

    def _get_one_column_type(self, label):
        """Get the column type of one column
        """
        if label in self._obj.columns:
            try:
                d = self._obj._column_types
            except AttributeError:
                return ColumnType.UNKNOWN
            try:
                return d[label]
            except KeyError:
                return ColumnType.UNKNOWN
        else:
            raise KeyError(str(label))

    def __getitem__(self, ind):
        if isinstance(ind, str):
            return self._get_one_column_type(ind)
        else:
            try:
                return [self._get_one_column_type(label) for label in ind]
            except TypeError:
                return self._get_one_column_type(ind)


class LabelsByTypeIndexer:
    """Indexing a PharmDataFrame to get labels from ColumnTypes
    """
    def __init__(self, acc):
        self._acc = acc

    def _get_one_label(self, tp):
        labels = self._get_many_labels(tp)
        if len(labels) == 0:
            raise KeyError(str(tp))
        elif len(labels) > 1:
            raise KeyError('Did not expect more than one ' + str(tp))
        else:
            return labels

    def _get_many_labels(self, column_type):
        """ Will raise if no columns of the type exists
            Always returns a list of labels
        """
        return [label for label in self._acc._obj.columns
                if self._acc.column_type[label] == column_type]

    def __getitem__(self, tp):
        try:
            len(tp)
            labels = []
            for t in tp:
                if t.max_one:
                    labels.extend(self._get_one_label(t))
                else:
                    labels.extend(self._get_many_labels(t))
            return labels
        except TypeError:
            if tp.max_one:
                return self._get_one_label(tp)
            else:
                return self._get_many_labels(tp)


@pd.api.extensions.register_dataframe_accessor('pharmpy')
class DataFrameAccessor:
    def __init__(self, obj):
        self._obj = obj

    @property
    def column_type(self):
        return ColumnTypeIndexer(self._obj)

    @property
    def labels_by_type(self):
        return LabelsByTypeIndexer(self)

    @property
    def id_label(self):
        """ Return the label of the id column
        """
        return self.labels_by_type[ColumnType.ID][0]

    @property
    def ids(self):
        """ Return the ids in the dataset
        """
        return self._obj[self.id_label].unique()

    @property
    def time_varying_covariates(self):
        """ Return a list of labels for all time varying covariates
        """
        cov_labels = self.labels_by_type[ColumnType.COVARIATE]
        if len(cov_labels) == 0:
            return []
        else:
            time_var = self._obj.groupby(by=self.id_label)[cov_labels].nunique().gt(1).any()
            return list(time_var.index[time_var])

    @property
    def baselines(self):
        """ Baselines for each id.
            Baseline is taken to be the first row even if that has a missing value.
        """
        idlab = self.id_label
        return self._obj.groupby(idlab).nth(0)

    @property
    def covariate_baselines(self):
        """ Return a dataframe with baselines of all covariates for each id.
            Baseline is taken to be the first row even if that has a missing value.
        """
        covariates = self.labels_by_type[ColumnType.COVARIATE]
        idlab = self.id_label
        df = self._obj[covariates + [idlab]]
        return df.groupby(idlab).nth(0)

    def generate_path(self, path=None, force=False):
        """Generate the path of dataframe if written.
           If no path is supplied or does not contain a filename a name is created
           from the name property of the PharmDataFrame.
           Will not overwrite unless forced.
        """
        if path is None:
            path = Path("")
        else:
            path = Path(path)
        if not path or path.is_dir():
            try:
                filename = f'{self._obj.name}.csv'
            except AttributeError:
                raise ValueError('Cannot name data file as no path argument was supplied and '
                                 'the DataFrame has no name property.')
            path /= filename
        if not force and path.exists():
            raise FileExistsError(f'File at {path} already exists.')
        return path

    def write_csv(self, path=None, force=False):
        """Write PharmDataFrame to a csv file
           return path for the written file
        """
        path = self.generate_path(path, force)
        self._obj.to_csv(path, na_rep=pharmpy.data.conf.na_rep, index=False)
        return path



# -*- encoding: utf-8 -*-
"""
================
Execution Engine
================

Job creator for a :class:`~pharmpy.generic.Model` implementation.

The critical, non-agnostic and central unit to inherit (e.g.
:class:`~pharmpy.api_nonmem.execute.NONMEM7`). That implementation can be multiclassed dynamically
for :class:`~pharmpy.tool.Tool` implementations, if mere duck typing doesn't cut it.

.. note:: An :class:`.Engine` implementation is expected to implement all methods/attributes.

Definitions
-----------
"""

from .environment import SystemEnvironment
from .run_directory import RunDirectory


class Engine:
    """An execution engine (e.g. for NONMEM7) of a :class:`~pharmpy.generic.Model`.

    Is a model API attached to attribute :attr:`Model.execute <pharmpy.generic.Model.execute>`.

    Arguments:
        envir: Type of :class:`~.environment.Environment` to use.

    Default `envir` inherits (depending on the current OS) :class:`~.environment.SystemEnvironment`
    (direct execution).
    """

    def __init__(self, model, envir=None):
        self.model = model
        if envir:
            self.environment = envir
        else:
            self.environment = SystemEnvironment()

    async def evaluate(self, cwd=None):
        """Starts model evaluation and returns :class:`~.job.Job` object.

        Arguments:
            cwd: Directory to create run directory in (temporary if None).

        .. todo:: Make either this level, or one up (:class:`pharmpy.generic.Model` and
            :class:`pharmpy.tool.Tool`), emulate -model_subdir command of PsN. Estimating
            pheno_real.mod with no dir info but the *parent* (otherwise it must be a temp dir),
            should create 'pheno_real/' dir only if it doesn't already exist, and *then* create
            pheno_real/estimate_dir1/, pheno_real/evaluate_dir74/, etc. (be clear but not collide).
        """

        raise NotImplementedError()

    async def estimate(self, cwd=None):
        """Starts model estimation and returns :class:`~.job.Job` object.

        Arguments:
            cwd: Directory to create run directory in (temporary if None).
        """

        return await self._execute_model(cwd)

    @property
    def bin(self):
        """Path (to main binary)."""
        raise NotImplementedError

    @property
    def version(self):
        """Version (of main binary)."""
        raise NotImplementedError

    def get_commandline(self, task):
        """Returns a command line for performing a task.

        Arguments:
            task: Any task of {'evaluate', 'estimate'}.
        """
        raise NotImplementedError

    async def _execute_model(self, cwd):
        rundir = RunDirectory(cwd, self.model.path.stem)
        rundir.model = self.model
        command = rundir.model.execute.get_commandline('estimate')
        envir = rundir.model.execute.environment

        job = await envir.submit(command, rundir)
        rundir.bind_job(job)
        return rundir

    def __bool__(self):
        """True if this engine is ready for executing *now*."""
        return False



import random
from collections import namedtuple

import pytest

import pharmpy.plugins.nonmem.nmtran_parser
import pharmpy.plugins.nonmem.records


@pytest.fixture
def records():
    return pharmpy.plugins.nonmem.records


@pytest.fixture
def parser():
    return pharmpy.plugins.nonmem.nmtran_parser.NMTranParser()


@pytest.fixture
def create_record(nonmem):
    """Record creating method with some (general) logging/asserts"""

    def func(buf, fail=False):
        record = nonmem.records.factory.create_record(buf)
        print(str(record))
        if fail:
            assert record.name is None
        else:
            assert record.name.startswith(record.raw_name)
        if buf.startswith('$'):
            assert str(record) == buf
        else:
            assert str(record) == '$' + buf
        return record

    return func


# -- ONLY ThetaRecord ------------------------------------------------------------------------------


class _GeneratedTheta:
    """
    A generated theta (see :class:`RandomThetas` for random generation).

    Can "messy" print and validate against tree parse, a.k.a. post-processing feature creep...
    """

    _WS = '\t' + (' '*9)
    Values = namedtuple('Values', ['low', 'init', 'up'])

    __slots__ = ('num', 'num_str', 'fixed', 'n_thetas')

    def __init__(self, num, num_str, fixed, n_thetas):
        self.num = self.Values(*num)
        self.num_str = self.Values(*num_str)
        self.fixed = fixed
        self.n_thetas = n_thetas

    @classmethod
    def new(cls, **kwargs):
        """Alternative constructor; Creates GeneratedTheta object from (partial) kwargs."""
        args = dict(num=(None, None, None), num_str=(None, None, None), fixed=None, n_thetas=None)
        if 'num' not in kwargs:
            num = tuple(kwargs.pop(x, None) for x in ['lower', 'init', 'upper'])
            kwargs.update(num=num)
        args.update(**kwargs)
        return cls(**args)

    @classmethod
    def from_tree(cls, tree):
        """
        Alternative constructor; Generate from AttrTree.

        This is a 'basic' non-API implementation of postprocessing, for grammar-close lexer-parser
        testing.
        """
        def find_tokens(tree, rule, rule_token):
            nodes = filter(lambda x: x.rule == rule, tree.tree_walk())
            return list(map(lambda x: getattr(x, rule_token), nodes))

        params = []
        for param in tree.all('param'):
            single = param.all('single')
            multi = param.all('multi')
            assert (len(single) + len(multi)) == 1
            params += single + multi

        thetas = []
        for param in params:
            kw = dict(num=[], fixed=None, n_thetas=None)
            for rule in ['low', 'init', 'up']:
                tok = find_tokens(param, rule, 'NUMERIC')
                assert len(tok) <= 1
                kw['num'] += tok if tok else [None]

            fixed = param.all('fix')
            assert len(fixed) <= 1
            kw['fixed'] = bool(fixed)

            n_thetas = find_tokens(param, 'n_thetas', 'INT')
            assert len(n_thetas) <= 1
            if n_thetas:
                kw['n_thetas'] = int(n_thetas[0])

            thetas += [cls.new(**kw)]
        return thetas

    def __eq__(self, other):
        """
        Equivalence check.

        Attributes which are None on self won't count. Thus partial equivalence, when __str__
        randomizing drops some info, works. No, this whole system isn't the best but it is a clean
        test of lexer-parser (without API magic).
        """
        for key, val in self.num._asdict().items():
            if val and val != getattr(other.num, key):
                return False
        for attr in ['fixed', 'n_thetas']:
            val = getattr(self, attr, None)
            if val and val != getattr(other, attr, None):
                return False
        return True

    def __str__(self):
        low = self._lr_pad(self.num_str.low)
        init = self._lr_pad(self.num_str.init)
        high = self._lr_pad(self.num_str.up)
        fix = ''
        if self.fixed:
            fix = self._lr_pad(random.choice(['FIX', 'FIXED']))
        form = random.randrange(6)
        if form == 0:
            out = init + fix
        elif form == 1:
            out = '(%s,%s %s)' % (low, init, fix)
        elif form == 2:
            out = '(%s,%s,%s %s)' % (low, init, high, fix)
        elif form == 3:
            out = '(%s,%s,%s) %s' % (low, init, high, fix)
        elif form == 4:
            init_missing = random.choice(self._WS)*random.randrange(3)
            out = '(%s,%s,%s %s)' % (low, init_missing, high, fix)
        else:
            n_thetas = self._l_pad('x') + self._lr_pad(self.n_thetas)
            out = '(%s)%s' % (init, n_thetas)
        return out

    def __repr__(self):
        """Pretty format on some 'standard form' (mostly for good pytest diffs)."""
        fix = ' FIX' if self.fixed else ''
        vals = [getattr(self.num, a) or '' for a in ['low', 'init', 'up']]
        out = '(%s, %s, %s%s)' % (*vals, fix)
        if self.n_thetas is not None:
            out += ('x%s' % (self.n_thetas,))
        return out

    def _l_pad(self, obj):
        """Format obj via wrapping with (left) random whitespace padding."""
        if obj is None:
            return ''
        lpad = random.choice(self._WS)*random.randrange(5)
        return lpad + str(obj)

    def _lr_pad(self, obj):
        """Format obj via wrapping with (left + right) random whitespace padding."""
        if obj is None:
            return ''
        lpad = random.choice(self._WS)*random.randrange(5)
        rpad = random.choice(self._WS)*random.randrange(5)
        return lpad + str(obj) + rpad


@pytest.fixture
def GeneratedTheta():
    """Provide GeneratedTheta (helper) class to tests."""
    return _GeneratedTheta


@pytest.fixture
def RandomThetas(RandomData, GeneratedTheta):
    """Provide RandomTheta (helper) class to tests."""

    class _RandomThetas(RandomData):
        """Extension of :class:`RandomData` with new generator, for producing thetas."""

        def __init__(self, *a, **kw):
            super().__init__(*a, **kw)
            self._gen_str_num = self.str_num(N=3)

        # -- methods producing random generators ---------------------------
        def theta(self):
            """Returns (generator for) random thetas."""

            def f():
                # TODO: This is backwards. Why should GeneratedTheta keep string-representation?
                # Causes issues whereas RandomData.str_num needs to do float(s) to return 'correct'
                # value for equivalence check...
                num, _str = zip(*next(self._gen_str_num))
                fixed = random.getrandbits(1)
                n_thetas = random.randrange(1, 5)
                return GeneratedTheta(num, _str, fixed, n_thetas)

            return self._gen(f)

    return _RandomThetas



import numpy as np
import pandas as pd

from pharmpy.math import is_posdef, nearest_posdef, sample_truncated_joint_normal


def sample_from_covariance_matrix(model, modelfit_results=None, parameters=None,
                                  force_posdef_samples=None, force_posdef_covmatrix=False, n=1):
    """Sample parameter vectors using the covariance matrix

       if modelfit_results is not provided the results from the model will be used

       :param parameters: use to only sample a subset of the parameters. None means all
       :param force_posdef_samples: Set to how many iterations to do before forcing all
                                    samples to be positive definite. None is default and means
                                    never and 0 means always
       :return: a dataframe with one sample per row
    """
    if modelfit_results is None:
        modelfit_results = model.modelfit_results

    if parameters is None:
        parameters = list(modelfit_results.parameter_estimates.index)

    pe = modelfit_results.parameter_estimates[parameters]
    index = pe.index
    mu = pe.to_numpy()
    sigma = modelfit_results.covariance_matrix[parameters].loc[parameters].to_numpy()
    if not is_posdef(sigma):
        if force_posdef_covmatrix:
            sigma = nearest_posdef(sigma)
        else:
            raise ValueError("Uncertainty covariance matrix not positive-definite")
    parameter_summary = model.parameters.summary().loc[parameters]
    parameter_summary = parameter_summary[~parameter_summary['fix']]
    a = parameter_summary.lower.astype('float64').to_numpy()
    b = parameter_summary.upper.astype('float64').to_numpy()

    # reject non-posdef
    kept_samples = pd.DataFrame()
    remaining = n

    if force_posdef_samples == 0:
        force_posdef = True
    else:
        force_posdef = False

    i = 0
    while remaining > 0:
        samples = sample_truncated_joint_normal(mu, sigma, a, b, n=remaining)
        df = pd.DataFrame(samples, columns=index)
        if not force_posdef:
            selected = df[df.apply(model.random_variables.validate_parameters, axis=1,
                                   use_cache=True)]
        else:
            selected = df.transform(model.random_variables.nearest_valid_parameters, axis=1)
        kept_samples = pd.concat((kept_samples, selected))
        remaining = n - len(kept_samples)
        i += 1
        if not force_posdef and force_posdef_samples is not None and i >= force_posdef_samples:
            force_posdef = True

    return kept_samples.reset_index(drop=True)


def sample_individual_estimates(model, parameters=None, samples_per_id=100):
    """Sample individual estimates given their covariance.

       :param parameters: A list of a subset of parameters to sample. Default is None,
                          which means all.
       :return: Pool of samples in a DataFrame
    """
    ests = model.modelfit_results.individual_estimates
    covs = model.modelfit_results.individual_estimates_covariance
    if parameters is None:
        parameters = ests.columns
    ests = ests[parameters]
    samples = pd.DataFrame()
    for (idx, mu), sigma in zip(ests.iterrows(), covs):
        sigma = sigma[parameters].loc[parameters]
        sigma = nearest_posdef(sigma)
        id_samples = np.random.multivariate_normal(mu.values, sigma.values, size=samples_per_id)
        id_df = pd.DataFrame(id_samples, columns=ests.columns)
        id_df.index = [idx] * len(id_df)        # ID as index
        samples = pd.concat((samples, id_df))
    return samples



from pharmpy import Model


def test_create_symbol(testdata):
    model = Model(testdata / 'nonmem' / 'pheno_real.mod')
    symbol = model.create_symbol(stem='ETAT')
    assert symbol.name == 'ETAT1'



# -*- encoding: utf-8 -*-
"""
Execution Environment
=====================

Platform (e.g. Linux) & system (e.g. SLURM) implementation.

The cluster/local or OS etc. to start jobs on.

Definitions
-----------
"""

import asyncio
import concurrent.futures
import logging
import os

from .job import Job


class Environment:
    """Manages execution of an engine on a platform/system.

    Subclasses may support e.g. Windows, Linux, SLURM or SGE.
    """

    def __new__(cls, *args, **kwargs):
        if cls is Environment:
            cls = SystemEnvironment
        self = cls.__new__(*args, **kwargs)
        if self is None or not self.supported:
            raise NotImplementedError("Cannot instantiate %r on your system" % (cls.__name__,))
        else:
            return self

    def submit(self, command, cwd):
        """Starts job and returns Job object."""
        raise NotImplementedError

    @property
    def supported(self):
        """True if environment supported on current platform."""
        pass


class SystemEnvironment(Environment):
    """Manages system (direct subprocess) execution for an engine on some platform.

    Needs :mod:`asyncio` because a running event loop is required to monitor child processes.
    Execution machinery via :mod:`concurrent.futures` interface.

    Attributes:
        jobs: Submitted jobs.
        futures: Future promises of jobs, instances of :mod:`concurrent.futures.Future`.
        pool: Threading pool, instance of :class:`~concurrent.Futures.ThreadPoolExecutor`.
    """

    def __new__(cls, *args, **kwargs):
        if cls is SystemEnvironment:
            if os.name == 'nt':
                cls = WindowsSystemEnvironment
            else:
                cls = PosixSystemEnvironment
        self = object.__new__(cls)
        self.__init__(*args, **kwargs)
        return self

    def __init__(self, threads=None):
        self.jobs = list()
        self.futures = list()
        self.pool = concurrent.futures.ThreadPoolExecutor(threads)

    async def submit(self, command, cwd=None):
        """Submits *command* to run as subprocess in *cwd* working directory."""

        logger = logging.getLogger(__name__)

        asyncio.get_child_watcher()
        job = Job(command, cwd,  stdout=self._stdout_handle, stderr=self._stderr_handle,
                  callback=self._callback, keepends=False)

        future = self.pool.submit(job.run)
        logger.debug('Submitted job %r (future: %r): %r', job, future, self)

        self.jobs += [job]
        self.futures += [future]
        return job

    async def wait(self, timeout=None, poll=1):
        """Wait (block) for all jobs to complete (and stop accepting new ones).

        Arguments:
            timeout: Maximum wait time in seconds.
            poll: Wait between each status check of job.
        """

        coros = [job.wait(timeout, poll) for job in self.jobs]
        await asyncio.gather(*coros)
        self.pool.shutdown()
        self.pool = None

    @property
    def log(self):
        return logging.getLogger('%s.%s' % (__name__, self.__class__.__name__))

    def _stdout_handle(self, line):
        self.log.info(line)

    def _stderr_handle(self, line):
        self.log.warning(line)

    def _callback(self, job):
        if job.rc == 0:
            self.log.info('Finished job %r: %r', job, self)
        else:
            self.log.error('Failed (rc=%d) job %r: %r', job.rc, job, self)

    def __repr__(self):
        return '<%s %s>' % (self.__class__.__name__, hex(id(self)))


class PosixSystemEnvironment(SystemEnvironment):
    """:class:`SystemEnvironment` on Linux/MacOS."""

    @property
    def supported(self):
        return (os.name != 'nt')


class WindowsSystemEnvironment(SystemEnvironment):
    """:class:`SystemEnvironment` on Windows."""

    def __init__(self, *args, **kwargs):
        asyncio.set_event_loop_policy(asyncio.WindowsProactorEventLoopPolicy())
        return super().__init__(*args, **kwargs)

    @property
    def supported(self):
        return (os.name == 'nt')



"""
Generic NONMEM option record class.

Assumes 'KEY=VALUE' and does not support 'KEY VALUE'.
"""

from collections import OrderedDict, namedtuple

from pharmpy.parse_utils.generic import AttrTree

from .record import Record


class OptionRecord(Record):
    @property
    def option_pairs(self):
        """ Extract the key-value pairs
            If no value exists set it to None
            Can only handle cases where options are supposed to be unique
        """
        pairs = OrderedDict()
        for node in self.root.all('option'):
            if hasattr(node, 'KEY'):
                pairs[node.KEY] = node.VALUE
            else:
                pairs[node.VALUE] = None
        return pairs

    @property
    def all_options(self):
        """ Extract all options even if non-unique.
            returns a list of named two-tuples with key and value
        """
        Option = namedtuple('Option', ['key', 'value'])
        pairs = []
        for node in self.root.all('option'):
            if hasattr(node, 'KEY'):
                pairs += [Option(node.KEY, node.VALUE)]
            else:
                pairs += [Option(node.VALUE, None)]
        return pairs

    def set_option(self, key, new_value):
        """ Set the value of an option

            If option already exists replaces its value
            appends option at the end if it does not exist
            does not handle abbreviations yet
        """
        # If already exists update value
        last_option = None
        for node in self.root.all('option'):
            old_key = node.find('KEY')
            if str(old_key) == key:
                node.VALUE = new_value
                return
            last_option = node

        ws_node = AttrTree.create('ws', [{'WS_ALL': ' '}])
        option_node = self._create_option(key, new_value)
        # If no other options add first else add just after last option
        if last_option is None:
            self.root.children = [ws_node, option_node] + self.root.children
        else:
            new_children = []
            for node in self.root.children:
                new_children.append(node)
                if node is last_option:
                    new_children += [ws_node, option_node]
            self.root.children = new_children

    def _create_option(self, key, value=None):
        if value is None:
            node = AttrTree.create('option', [{'VALUE': key}])
        else:
            node = AttrTree.create('option', [{'KEY': key}, {'EQUAL': '='}, {'VALUE': value}])
        return node

    def append_option(self, key, value=None):
        """ Append option as last option

            Method applicable to option records with no special grammar
        """
        node = self._create_option(key, value)
        self.append_option_node(node)

    def append_option_node(self, node):
        """ Add a new option as last option
        """
        last_child = self.root.children[-1]
        if '\n' in str(last_child):
            if len(self.root.children) > 1 and self.root.children[-2].rule == 'ws':
                self.root.children[-1:0] = [node]
            else:
                ws_node = AttrTree.create('ws', [{'WS_ALL': ' '}])
                self.root.children[-1:0] = [ws_node, node]
        else:
            self.root.children.append(node)



# import asyncio
# import logging
# import threading

# import pytest

# from pharmpy.execute.job import Job


# class JobAssertions:
#    def __init__(self, stdout, stderr):
#        logging.getLogger().debug('Initializing: %r', self)
#        self.stdout_reference = stdout
#        self.stderr_reference = stderr
#        self.stdout_captured = []
#        self.stderr_captured = []

#    def stdout(self, line):
#        with threading.Lock():
#            self.stdout_captured += [line]

#    def stderr(self, line):
#        with threading.Lock():
#            self.stderr_captured += [line]

#    def callback(self, job):
#        logging.getLogger().debug('Callback (rc=%r) from job=%r: %r', job.rc, job, self)
#        with threading.Lock():
#            self.job = job

#    def assert_captures(self):
#        if hasattr(self, 'job'):
#            logging.getLogger().debug('Asserting job (rc=%r) %r: %r', self.job.rc, self.job, self)
#        else:
#            raise RuntimeError('No job has called back: %r' % self)
#        with threading.Lock():
#            assert self.job.rc == 0
#            assert self.stdout_reference == self.stdout_captured
#            assert self.stderr_reference == self.stderr_captured

#    def __repr__(self):
#        with threading.Lock():
#            return '<%s %s>' % (self.__class__.__name__, hex(id(self)))


# @pytest.mark.asyncio
# async def test_job_start_single(py_command):
#    command = py_command[0]
#    asyncio.get_child_watcher()

#    job = Job(command)
#    job.start()
#    job.init.wait()

#    await job.wait(3, 0.1)
#    assert job.init
#    assert job.done
#    assert job.rc == 0


# def test_job_start_series(py_command_slow, event_loop):
#    """Test running job completely asynchronously."""

#    command, output = py_command_slow

#    async def test_one():
#        watcher = asyncio.get_child_watcher()
#        watcher.attach_loop(asyncio.get_event_loop())

#        ref = JobAssertions(stdout=output['stdout'], stderr=output['stderr'])
#        job = Job(command, stdout=ref.stdout, stderr=ref.stderr, callback=ref.callback)
#        job.start()
#        assert not job.init

#        job.init.wait()
#        assert not job.done
#        assert job.rc is None
#        assert job.proc.pid > 0

#        await job.wait(3, 0.1)
#        assert job.init
#        assert job.done
#        ref.assert_captures()

#    async def test_series():
#        for i in range(5):
#            await test_one()

#    event_loop.run_until_complete(test_series())


# @pytest.mark.asyncio
# async def test_job_start_parallel(py_command_slow):
#    """Test running job completely asynchronously."""

#    command, output = py_command_slow
#    asyncio.get_child_watcher()

#    jobs, refs = [], []
#    for i in range(5):
#        ref = JobAssertions(stdout=output['stdout'], stderr=output['stderr'])
#        job = Job(command, stdout=ref.stdout, stderr=ref.stderr, callback=ref.callback)
#        jobs += [job]
#        refs += [ref]
#        job.start()
#        assert not job.init

#    for job in jobs:
#        job.init.wait()
#        assert not job.done

#    for ref, job in zip(refs, jobs):
#        await job.wait(3, 0.1)
#        ref.assert_captures()



"""
=======
Pharmpy
=======

Pharmpy is a python package for pharmacometrics modelling.

Definitions
===========
"""

__version__ = '0.3.0'

import logging

from .model_factory import Model
from .parameter import Parameter, ParameterSet

logging.getLogger(__name__).addHandler(logging.NullHandler())

__all__ = ['Model', 'Parameter', 'ParameterSet']



# The pharmpy visualization module

# Since the python visualization and plotting landscape is rapidly
# evolving and there are many different modules to choose from
# all visualization API calls should be made from this module so
# that we could start using another API more easily.
# Design conciderations:
# We would like to be able to have interactive plots which currently
# means to select a package that can render to html. The two main
# contenders that can do this are the altair and the bokeh libraries.
# Bokeh seems to have a larger community, but the strength in altair
# is its use of the standard vega-lite format, which decouples the
# creation of a plot from the rendering. Altair plots (or rather vega
# plots) can be changed from the html directly via the online vega
# editor. So using altair for now, but expecting to revisit this
# decision shortly.

# Will provide base functions for creating different types of plots
# or other types of visualizations

import altair as alt
import pandas as pd

_chart_width = 500
_chart_height = 500


def scatter_plot_correlation(df, x, y, title=""):
    chart = alt.Chart(df, width=_chart_width, height=_chart_height).mark_circle(size=100).encode(
        alt.X(x),
        alt.Y(y),
        tooltip=[x, y]
    ).properties(
        title=title,
    ).interactive()

    line = alt.Chart(
        pd.DataFrame({x: [min(df[x]), max(df[x])],
                      y: [min(df[y]), max(df[y])]})).mark_line().encode(
            alt.X(x),
            alt.Y(y),
    ).interactive()

    plot = chart + line

    plot = plot.configure_title(fontSize=16)
    plot = plot.configure_axis(labelFontSize=12, titleFontSize=14)
    return plot


def histogram(values, title=""):
    """Histogram with percentage on y and a rule at mean
       slider for reducing the number of values used.
    """
    df = pd.DataFrame({values.name: values, 'num': list(range(1, len(values) + 1))})

    slider = alt.binding_range(min=1, max=len(values), step=1, name='Number of samples: ')
    selection = alt.selection_single(bind=slider, fields=['num'], name="num",
                                     init={'num': len(values)})

    base = alt.Chart(df).transform_filter('datum.num <= num_num')

    plot = base.transform_joinaggregate(
        total='count(*)'
    ).transform_calculate(
        pct='1 / datum.total'
    ).mark_bar().encode(
        alt.X(f'{values.name}:Q', bin=True),
        alt.Y('sum(pct):Q', axis=alt.Axis(format='%'))
    ).add_selection(
        selection
    ).properties(
        title=title
    )

    rule = base.mark_rule(color='red').encode(
        x=f'mean({values.name}):Q',
        size=alt.value(5)
    )

    return plot + rule



"""
===================
Generic Model class
===================

**Base class of all implementations.**

Inherit to *implement*, i.e. to define support for a specific model type. Duck typing is utilized,
but an implementation is expected to implement **all** methods/attributes.

Definitions
-----------
"""

import copy
from pathlib import Path

import sympy


class ModelException(Exception):
    pass


class ModelSyntaxError(ModelException):
    def __init__(self, msg='model syntax error'):
        super().__init__(msg)


class Model:
    """
     Property: name
    """
    @property
    def modelfit_results(self):
        return None

    def update_source(self):
        """Update the source"""
        self.source.code = str(self)

    def write(self, path='', force=False):
        """Write model to file using its source format
           If no path is supplied or does not contain a filename a name is created
           from the name property of the model
           Will not overwrite in case force is True.
           return path written to
        """
        path = Path(path)
        if not path or path.is_dir():
            try:
                filename = f'{self.name}{self.source.filename_extension}'
            except AttributeError:
                raise ValueError('Cannot name model file as no path argument was supplied and the'
                                 'model has no name.')
            path = path / filename
        if not force and path.exists():
            raise FileExistsError(f'File {path} already exists.')
        self.update_source(path=path, force=force)
        self.source.write(path, force=force)
        return path

    def update_inits(self):
        """Update inital estimates of model from its own ModelfitResults
        """
        if self.modelfit_results:
            self.parameters = self.modelfit_results.parameter_estimates
        else:
            # FIXME: Other exception here. ModelfitError?
            raise ModelException("Cannot update initial parameter estimates "
                                 "since parameters were not estimated")

    def copy(self):
        """Create a deepcopy of the model object"""
        return copy.deepcopy(self)

    def update_individual_estimates(self, source):
        self.initial_individual_estimates = self.modelfit_results.individual_estimates

    @property
    def dataset(self):
        raise NotImplementedError()

    def read_raw_dataset(self, parse_columns=tuple()):
        raise NotImplementedError()

    def create_symbol(self, stem):
        """Create a new unique variable symbol

           stem - First part of the new variable name
        """
        # TODO: Also check parameter and rv names and dataset columns
        symbols = self.statements.free_symbols
        i = 1
        while True:
            candidate = sympy.Symbol(f'{stem}{i}')
            if candidate not in symbols:
                return candidate
            i += 1



# import dateutil.parser
import re

# Warning! Bad and experimental code below.
# Should ideally be able to parse on demand or at least be able to skip sections


class NONMEMResultsFile:
    '''Representing and parsing a NONMEM results file (aka lst-file)
       This is not a generic output object and will be combined by other classes
       into new structures. For example if ext file exists we would not use the
       estimates from the lst-file.
    '''
    def __init__(self, path):
        with open(str(path), 'r') as res_file:
            content = res_file.readlines()
        parts = self._split_content(content)
        self._parse_datestamps(parts['start_timestamp'])
        self.model_code = parts['model_code']
        self.nmtran_messages = parts['nmtran_messages']
        self._parse_version(parts['about_text'])

    def _split_content(self, content):
        '''First coarse parser. Splitting the file in parts that will be parsed separately
        '''
        parts = {}
        parts['start_timestamp'] = content.pop(0)

        (parts['model_code'], content) = self._split_until_regexp(
                content, ['NM-TRAN MESSAGES', 'License Registered to:'], remove_lines=1)

        if content[0].startswith('NM-TRAN MESSAGES'):
            (parts['nmtran_messages'], content) = \
                    self._split_until_regexp(content, ['License Registered to:'], remove_lines=1)

        (_, content) = self._split_until_regexp(content, ['1NONLINEAR MIXED EFFECTS MODEL PROGRAM'])
        (parts['about_text'], content) = self._split_until_regexp(content, [' PROBLEM NO.:'])
        return parts

    def _split_until_regexp(self, content, regexps, remove_lines=0):
        '''Split an array of strings into two parts. One before the match of any of the regexps and one
        after and including the match.
        remove_lines is the number of lines to remove at the end of the before part
        '''
        match = False
        for i, line in enumerate(content):
            for regexp in regexps:
                if re.match(regexp, line):
                    match = True
            if match:
                break
        else:
            raise EOFError
        return (content[0:(i - remove_lines)], content[i:])

    def _parse_version(self, about_text):
        m = re.match(r'1NONLINEAR MIXED EFFECTS MODEL PROGRAM \(NONMEM\) VERSION (.*)\n',
                     about_text[0])
        if m:
            self.nonmem_version = m.group(1)

    def _parse_datestamps(self, raw_string):
        pass  # Crap! This dateutil doesn't seem to support Swedish
        # Had a look at dateparser that could support Swedish, but it needs patching for that
        # self.start_timestamp = dateutil.parser.parse(raw_string)



from toil.common import Toil
from toil.job import Job

from pharmpy.data.iterators import Resample
from pharmpy.results import BootstrapResults

from .jobs import InputModel, NonmemJob

toil_options = Job.Runner.getDefaultOptions("file:jobStore")
toil_options.clean = "always"


class BootstrapPrepare(Job):
    """Prepare all bootstrap models and datasets
    """
    def __init__(self, input_model, resamples=10):
        super().__init__()
        self.input_model = input_model
        self.resamples = resamples

    def run(self, file_store):
        resampler = Resample(self.input_model.model, group='ID', resamples=self.resamples,
                             replace=True)
        rvs = []
        for model, _ in resampler:
            job = NonmemJob(model, [])  # No extra files since dataset is recreated
            child = self.addChild(job)
            rvs.append(child.rv())
        return self.addFollowOn(BootstrapPostprocess(self.input_model.model, rvs)).rv()


class BootstrapPostprocess(Job):
    def __init__(self, original_model, rvs):
        super().__init__()
        self.original_model = original_model
        self.rvs = rvs

    def run(self, file_store):
        res = BootstrapResults(self.original_model, self.rvs)
        return res


class BootstrapWorkflow:
    """Workflow to perform a bootstrap
    """
    def __init__(self, model):
        self._model = model

    def start(self):
        with Toil(toil_options) as workflow:
            input_model = InputModel(workflow, self._model)
            job = BootstrapPrepare(input_model)
            res = workflow.start(job)
            print(res.parameter_estimates)



from toil.common import Toil
from toil.job import Job

from .jobs import BatchModelfit, InputModel, export_files

toil_options = Job.Runner.getDefaultOptions("file:jobStore")
toil_options.clean = "always"


class ModelfitWorkflow:
    """Workflow for a simple modelfit

       NONMEM only for now, but will become the agnostic workflow
       Can handle multiple models
       Copies up lst and ext files
    """
    def __init__(self, models):
        self._models = models

    def start(self):
        with Toil(toil_options) as workflow:
            input_models = [InputModel(workflow, model) for model in self._models]
            job = BatchModelfit(input_models)
            result_models = workflow.start(job)
            for model in result_models:
                export_files(workflow, model.modelfit_results.tool_files)



"""

.. list-table:: Options for the results module
   :widths: 25 25 50 150
   :header-rows: 1

   * - Option name
     - Default value
     - Type
     - Description
   * - ``native_shrinkage``
     - ``True``
     - bool
     - Should shrinkage calculation of external tool be used.
       Otherwise pharmpy will calculate shrinkage
"""

import json
from pathlib import Path

import altair as alt
import numpy as np
import pandas as pd

import pharmpy.config as config
import pharmpy.visualization
from pharmpy.data import PharmDataFrame
from pharmpy.math import cov2corr


class ResultsConfiguration(config.Configuration):
    native_shrinkage = config.ConfigItem(True, 'Use shrinkage results from external tool')


conf = ResultsConfiguration()


class ResultsJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, pd.DataFrame) or isinstance(obj, pd.Series):
            d = json.loads(obj.to_json(orient='table'))
            if isinstance(obj, PharmDataFrame):
                d['__class__'] = 'DataFrame'
            else:
                d['__class__'] = obj.__class__.__name__
            return d
        elif obj.__class__.__module__.startswith('altair.'):
            d = obj.to_dict()
            d['__class__'] = 'vega-lite'
            return d
        else:
            return json.JSONEncoder.encode(self, obj)


class ResultsJSONDecoder(json.JSONDecoder):
    def __init__(self, *args, **kwargs):
        json.JSONDecoder.__init__(self, object_hook=self.object_hook, *args, **kwargs)
        self.cls = None

    def object_hook(self, dct):
        if '__class__' in dct:
            cls = dct['__class__']
            del dct['__class__']
            if cls == 'DataFrame' or cls == 'Series':
                res = pd.read_json(json.dumps(dct), orient='table')
                return res
            elif cls == 'vega-lite':
                res = alt.Chart.from_dict(dct)
            else:
                self.cls = cls
        return dct


def read_results(path_or_buf):
    try:
        path = Path(path_or_buf)
        if not path.is_file():
            raise FileNotFoundError
    except (FileNotFoundError, OSError, TypeError, ValueError):
        s = path_or_buf
    else:
        with open(path, 'r') as json_file:
            s = json_file.read()
    decoder = ResultsJSONDecoder()
    d = decoder.decode(s)
    if decoder.cls == 'FREMResults':
        from pharmpy.methods.frem import FREMResults
        res = FREMResults.from_dict(d)
    elif decoder.cls == 'BootstrapResults':
        from pharmpy.methods.bootstrap import BootstrapResults
        res = BootstrapResults(original_model=None, bootstrap_models=None)
        res._statistics = d['statistics']

    return res


class Results:
    """ Base class for all result classes
    """
    def to_json(self, path=None):
        json_dict = self.to_dict()
        json_dict['__class__'] = self.__class__.__name__
        s = json.dumps(json_dict, cls=ResultsJSONEncoder)
        if path:
            with open(path, 'w') as fh:
                fh.write(s)
        else:
            return s

    def to_dict(self):
        """Convert results object to a dictionary
        """
        raise NotImplementedError()

    @classmethod
    def from_dict(cls, d):
        """Cread a results object from a dictionary
        """
        raise NotImplementedError()

    def to_csv(self, path):
        """Save results as a human readable csv file

           Index will not be printed if it is a basic range.
        """
        d = self.to_dict()
        s = ""
        for key, value in d.items():
            if value.__class__.__module__.startswith('altair.'):
                continue
            s += f'{key}\n'
            if isinstance(value, pd.DataFrame):
                if isinstance(value.index, pd.RangeIndex):
                    use_index = False
                else:
                    use_index = True
                s += value.to_csv(index=use_index)
            else:
                s += str(value)
            s += '\n'
        with open(path, 'w') as fh:
            print(s, file=fh)

    def create_report(self, path):
        import pharmpy.reporting.reporting as reporting
        reporting.generate_report(self.rst_path, path)

    def add_plots(self):
        """Create and add all plots to results object
        """
        raise NotImplementedError()


class ModelfitResults:
    """ Base class for results from a modelfit operation

    properties: individual_OFV is a df with currently ID and iOFV columns
        model_name - name of model that generated the results
        model
    """
    def __init__(self, ofv=None, parameter_estimates=None, covariance_matrix=None,
                 standard_errors=None):
        self._ofv = ofv
        self._parameter_estimates = parameter_estimates
        self._covariance_matrix = covariance_matrix
        self._standard_errors = standard_errors

    def reparameterize(self, parameterizations):
        """Reparametrize all parameters given a list of parametrization object
           will change the parameter_estimates and standard_errors to be for
           the transformed parameter
        """
        raise NotImplementedError("Not implemented")

    @property
    def ofv(self):
        """Final objective function value
        """
        return self._ofv

    @property
    def parameter_estimates(self):
        """Parameter estimates as series
        """
        return self._parameter_estimates

    @property
    def covariance_matrix(self):
        """The covariance matrix of the population parameter estimates
        """
        return self._covariance_matrix

    def _cov_from_inf(self):
        Im = self.information_matrix
        cov = pd.DataFrame(np.linalg.inv(Im.values), index=Im.index, columns=Im.columns)
        return cov

    def _cov_from_corrse(self):
        se = self.standard_errors
        corr = self.correlation_matrix
        x = se.values @ se.values.T
        cov = x * corr.values
        cov_df = pd.DataFrame(cov, index=corr.index, columns=corr.columns)
        return cov_df

    @property
    def information_matrix(self):
        """The Fischer information matrix of the population parameter estimates
        """
        raise NotImplementedError()

    def _inf_from_cov(self):
        C = self.covariance_matrix
        Im = pd.DataFrame(np.linalg.inv(C.values), index=C.index, columns=C.columns)
        return Im

    def _inf_from_corrse(self):
        se = self.standard_errors
        corr = self.correlation_matrix
        x = se.values @ se.values.T
        cov = x * corr.values
        Im = pd.DataFrame(np.linalg.inv(cov), index=corr.index, columns=corr.columns)
        return Im

    @property
    def correlation_matrix(self):
        """The correlation matrix of the population parameter estimates
        """
        raise NotImplementedError()

    def _corr_from_cov(self):
        C = self.covariance_matrix
        corr = pd.DataFrame(cov2corr(C.values), index=C.index, columns=C.columns)
        return corr

    def _corr_from_inf(self):
        Im = self.information_matrix
        corr = pd.DataFrame(cov2corr(np.linalg.inv(Im.values)), index=Im.index, columns=Im.columns)
        return corr

    @property
    def standard_errors(self):
        """Standard errors of population parameter estimates
        """
        return self._standard_errors

    @property
    def relative_standard_errors(self):
        """Relative standard errors of popilation parameter estimates
        """
        ser = self.standard_errors / self.parameter_estimates
        ser.name = 'RSE'
        return ser

    def _se_from_cov(self):
        """Calculate the standard errors from the covariance matrix
           can be used by subclasses
        """
        cov = self.covariance_matrix
        se = pd.Series(np.sqrt(np.diag(cov.values)), index=cov.index)
        return se

    def _se_from_inf(self):
        """Calculate the standard errors from the information matrix
        """
        Im = self.information_matrix
        se = pd.Series(np.sqrt(np.linalg.inv(Im.values)), index=Im.index)
        return se

    @property
    def individual_ofv(self):
        """A Series with individual estimates indexed over ID
        """
        raise NotImplementedError("Not implemented")

    @property
    def individual_estimates(self):
        """Individual parameter estimates

           A DataFrame with ID as index one column for each individual parameter
        """
        raise NotImplementedError("Not implemented")

    @property
    def individual_estimates_covariance(self):
        """The covariance matrix of the individual estimates
        """
        raise NotImplementedError("Not implemented")

    def eta_shrinkage(self, sd=False):
        """Eta shrinkage for each eta

           Variance = False to get sd scale
        """
        pe = self.parameter_estimates
        # Want parameter estimates combined with fixed parameter values
        param_inits = self.model.parameters.summary()['value']
        pe = pe.combine_first(param_inits)

        ie = self.individual_estimates
        parameters = self.model.random_variables.iiv_variance_parameters()
        param_names = [param.name for param in parameters]
        diag_ests = pe[param_names]
        diag_ests.index = ie.columns
        if not sd:
            shrinkage = 1 - (ie.var() / diag_ests)
        else:
            shrinkage = 1 - (ie.std() / (diag_ests ** 0.5))
        return shrinkage

    @property
    def individual_shrinkage(self):
        """The individual eta-shrinkage

            Definition: ieta_shr = (var(eta) / omega)
        """
        cov = self.individual_estimates_covariance
        pe = self.parameter_estimates
        # Want parameter estimates combined with fixed parameter values
        param_inits = self.model.parameters.summary()['value']
        pe = pe.combine_first(param_inits)

        # Get all iiv variance parameters
        parameters = self.model.random_variables.iiv_variance_parameters()
        param_names = [param.name for param in parameters]
        diag_ests = pe[param_names]

        def fn(row, ests):
            names = row[0].index
            ser = pd.Series(np.diag(row[0].values) / ests, index=names)
            return ser

        ish = pd.DataFrame(cov).apply(fn, axis=1, ests=diag_ests.values)
        return ish

    def parameter_summary(self):
        """Summary of parameter estimates and uncertainty
        """
        pe = self.parameter_estimates
        ses = self.standard_errors
        rses = self.relative_standard_errors
        df = pd.DataFrame({'estimate': pe, 'SE': ses, 'RSE': rses})
        return df

    def plot_iofv_vs_iofv(self, other):
        x_label = f'{self.model_name} iOFV'
        y_label = f'{other.model_name} iOFV'
        df = PharmDataFrame({x_label: self.individual_ofv, y_label: other.individual_ofv})
        plot = pharmpy.visualization.scatter_plot_correlation(df, x_label, y_label,
                                                              title='iOFV vs iOFV')
        return plot


class ChainedModelfitResults(list, ModelfitResults):
    """A list of modelfit results given in order from first to final
       inherits from both list and ModelfitResults. Each method from ModelfitResults
       will be performed on the final modelfit object
    """
    def reparameterize(self, parameterizations):
        return self[-1].reparameterize(parameterizations)

    @property
    def ofv(self):
        return self[-1].ofv

    @property
    def parameter_estimates(self):
        return self[-1].parameter_estimates

    @property
    def covariance_matrix(self):
        return self[-1].covariance_matrix

    @property
    def information_matrix(self):
        return self[-1].information_matrix

    @property
    def correlation_matrix(self):
        return self[-1].correlation_matrix

    @property
    def standard_errors(self):
        return self[-1].standard_errors

    @property
    def individual_ofv(self):
        return self[-1].individual_ofv

    @property
    def individual_estimates(self):
        return self[-1].individual_estimates

    @property
    def individual_estimates_covariance(self):
        return self[-1].individual_estimates_covariance

    @property
    def individual_shrinkage(self):
        return self[-1].individual_shrinkage

    def plot_iofv_vs_iofv(self, other):
        return self[-1].plot_iofv_vs_iofv(other)

    @property
    def model_name(self):
        return self[-1].model_name

    # FIXME: To not have to manually intercept everything here. Could do it in a general way.



# import sys

# import pytest

# from pharmpy.execute.environment import SystemEnvironment


# @pytest.mark.asyncio
# async def test_SystemEnvironment_submit_series(py_command):
#    """Test SystemEnvironment.submit, single job."""

#    command, output = py_command

#    envir = SystemEnvironment(threads=4)
#    assert not envir.jobs
#    assert not envir.futures

#    for i in range(5):
#        await envir.submit(command=command)
#        assert len(envir.jobs) == (i+1)
#        assert len(envir.futures) == (i+1)

#        job = envir.jobs[0]
#        await job.wait(1, 0.1)
#        assert job.init
#        assert job.done
#        assert job.rc == 0
#        assert job.proc.pid > 0


# @pytest.mark.asyncio
# async def test_SystemEnvironment_submit_parallel():
#    """Test SystemEnvironment.submit, many jobs."""

#    envir = SystemEnvironment(threads=4)
#    assert not envir.jobs
#    assert not envir.futures

#    commands = [
#        [sys.executable, '-c', "import time; time.sleep(0.20); print('OUT[5]')"],
#        [sys.executable, '-c', "import time; time.sleep(0.15); print('OUT[4]')"],
#        [sys.executable, '-c', "import time; time.sleep(0.10); print('OUT[3]')"],
#        [sys.executable, '-c', "import time; time.sleep(0.05); print('OUT[2]')"],
#        [sys.executable, '-c', "print('OUT[1]')"],
#    ]

#    for i, command in enumerate(commands):
#        await envir.submit(command=command, cwd=None)
#        assert len(envir.jobs) == (i+1)
#        assert len(envir.futures) == (i+1)
#        assert not envir.jobs[i].init
#        assert not envir.jobs[i].done

#    await envir.wait(1, 0.1)
#    for job in envir.jobs:
#        assert job.init
#        assert job.done
#        assert job.rc == 0
#        assert job.proc.pid > 0



import numpy as np
import pandas as pd
import pytest

from pharmpy import Model
from pharmpy.parameter_sampling import sample_from_covariance_matrix, sample_individual_estimates


def test_sample_from_covariance_matrix(testdata):
    model = Model(testdata / 'nonmem' / 'pheno_real.mod')
    np.random.seed(318)
    samples = sample_from_covariance_matrix(model, n=3)
    correct = pd.DataFrame({'THETA(1)': [0.004965, 0.004811, 0.004631],
                            'THETA(2)': [0.979979, 1.042210, 0.962791],
                            'THETA(3)': [0.007825, -0.069350, 0.052367],
                            'OMEGA(1,1)': [0.019811, 0.059127, 0.030619],
                            'OMEGA(2,2)': [0.025248, 0.029088, 0.019749],
                            'SIGMA(1,1)': [0.014700, 0.014347, 0.011470]})
    pd.testing.assert_frame_equal(samples, correct, check_less_precise=True)


def test_sample_individual_estimates(testdata):
    model = Model(testdata / 'nonmem' / 'pheno_real.mod')
    np.random.seed(86)
    samples = sample_individual_estimates(model)
    assert len(samples) == 59 * 100
    assert list(samples.columns) == ['ETA(1)', 'ETA(2)']
    assert pytest.approx(samples.iloc[0]['ETA(1)'], 1e-5) == 0.0418399
    assert pytest.approx(samples.iloc[0]['ETA(2)'], 1e-5) == -0.0587623

    restricted = sample_individual_estimates(model, parameters=['ETA(2)'], samples_per_id=1)
    assert len(restricted) == 59
    assert restricted.columns == ['ETA(2)']
    assert pytest.approx(restricted.iloc[0]['ETA(2)'], 1e-5) == -0.0167246



import pandas as pd

import pharmpy.visualization
from pharmpy.results import Results


class BootstrapResults(Results):
    # FIXME: Could inherit from results that take multiple runs like bootstrap, cdd etc.
    def __init__(self, original_model, bootstrap_models):
        if original_model is None and bootstrap_models is None:
            # FIXME: this is a special case for now for json handling before we have modelfit json
            return
        self._original_results = original_model.modelfit_results
        self._bootstrap_results = [m.modelfit_results for m in bootstrap_models
                                   if m.modelfit_results is not None]
        self._total_number_of_models = len(bootstrap_models)

    @property
    def ofv(self):
        boot_ofvs = [x.ofv for x in self._bootstrap_results]
        return pd.Series(boot_ofvs, name='ofv')

    @property
    def parameter_estimates(self):
        df = pd.DataFrame()
        for res in self._bootstrap_results:
            df = df.append(res.parameter_estimates, ignore_index=True, sort=False)
        df = df.reindex(self._bootstrap_results[0].parameter_estimates.index, axis=1)
        df = df.reset_index(drop=True)
        return df

    @property
    def standard_errors(self):
        pass
        # FIXME: Continue here

    @property
    def statistics(self):
        try:
            return self._statistics
        except AttributeError:
            pass
        df = self.parameter_estimates
        ofvs = self.ofv
        df.insert(0, 'OFV', ofvs)
        orig = self._original_results.parameter_estimates
        ofv_ser = pd.Series({'OFV': self._original_results.ofv})
        orig = pd.concat([ofv_ser, orig])
        mean = df.mean()
        bias = mean - orig
        summary = pd.DataFrame({'mean': mean, 'bias': bias, 'stderr': df.std()})
        self._statistics = summary
        return summary

    @property
    def distribution(self):
        try:
            return self._distribution
        except AttributeError:
            pass
        df = self.parameter_estimates
        ofvs = self.ofv
        df.insert(0, 'OFV', ofvs)
        summary = pd.DataFrame({'min': df.min(), '0.05%': df.quantile(0.0005),
                                '0.5%': df.quantile(0.005), '2.5%': df.quantile(0.025),
                                '5%': df.quantile(0.05), 'median': df.median(),
                                '95%': df.quantile(0.95), '97.5%': df.quantile(0.975),
                                '99.5%': df.quantile(0.995), '99.95%': df.quantile(0.9995),
                                'max': df.max()})
        self._distribution = summary
        return summary

    def __repr__(self):
        inclusions = f'Inclusion\n\nTotal number of models: {self._total_number_of_models}\n' \
                     f'Models not included because of failure: ' \
                     f'{self._total_number_of_models - len(self._bootstrap_results)}\n' \
                     f'Models included in analysis: {len(self._bootstrap_results)}'
        statistics = f'Statistics\n{repr(self.statistics)}'
        distribution = f'Distribution\n{repr(self.distribution)}'
        return f'{inclusions}\n\n{statistics}\n\n{distribution}'

    def to_dict(self):
        return {'statistics': self.statistics}

    def plot_ofv(self):
        plot = pharmpy.visualization.histogram(self.ofv, title='Bootstrap OFV')
        return plot



# -*- encoding: utf-8 -*-

from pharmpy.execute.run_directory import RunDirectory


def test_RunDirectory_temporary():
    rd = RunDirectory()
    path = rd.path
    assert path.is_dir()

    del rd
    assert not path.exists()


def test_RunDirectory_persistent():
    rd = RunDirectory('.')
    path = rd.path
    assert path.is_dir()

    contents = list(rd.path.iterdir())
    assert not contents
    rd.cleanlevel = 1
    rd.clean_config(level=1, patterns=['*.temp'], rm_dirs=False)
    to_remove = path / 'some.temp'
    to_keep = path / 'save.txt'
    open(str(to_remove), 'a').close()
    open(str(to_keep), 'a').close()

    rd.cleanup()
    assert path.exists()
    assert to_keep.exists()
    assert not to_remove.exists()

    del rd
    assert path.exists()
    assert to_keep.exists()
    (path / 'save.txt').unlink()
    path.rmdir()

    assert not path.exists()



import pandas as pd

from pharmpy.methods.bootstrap import BootstrapResults
from pharmpy.results import ModelfitResults, read_results


def test_bootstrap():
    orig = ModelfitResults(ofv=110, parameter_estimates=pd.Series({'TVCL': 0.75, 'TVV': 2.25}))
    res1 = ModelfitResults(ofv=100, parameter_estimates=pd.Series({'TVCL': 1.0, 'TVV': 2.0}))
    res2 = ModelfitResults(ofv=120, parameter_estimates=pd.Series({'TVCL': 1.5, 'TVV': 3.0}))

    class MyModel:
        pass
    orig_mod = MyModel()
    orig_mod.modelfit_results = orig
    res1_mod = MyModel()
    res1_mod.modelfit_results = res1
    res2_mod = MyModel()
    res2_mod.modelfit_results = res2

    boot = BootstrapResults(orig_mod, [res1_mod, res2_mod])
    correct_statistics = pd.DataFrame({'mean': [110.0, 1.25, 2.5],
                                       'bias': [0.0, 0.5, 0.25],
                                       'stderr': [14.142136, 0.353553, 0.707107]},
                                      index=['OFV', 'TVCL', 'TVV'])
    pd.testing.assert_frame_equal(boot.parameter_estimates, pd.DataFrame({'TVCL': [1.0, 1.5],
                                                                          'TVV': [2.0, 3.0]}))
    pd.testing.assert_frame_equal(boot.statistics, correct_statistics)

    # json round trip
    json = boot.to_json()
    obj = read_results(json)
    pd.testing.assert_frame_equal(obj.statistics, correct_statistics)



# -*- encoding: utf-8 -*-
"""
============
Executed Job
============

A job unit, for non-blocking execution and communication (of subprocesses).

Can be used raw for running subprocesses asynchronously.

Definitions
-----------
"""

import asyncio
import functools
import locale
import logging
import os
import queue
import threading
import time
from asyncio.subprocess import PIPE


class Status(threading.Event):
    """A started/done status of a :class:`~Job` instance."""

    def __bool__(self):
        return self.is_set()

    def __repr__(self):
        return '<%s %r>' % (self.__class__.__name__, bool(self))


class Job(threading.Thread):
    """A job of an :class:`~pharmpy.execute.Engine` running in a
    :class:`~pharmpy.execute.RunDirectory` (via an :class:`~pharmpy.execute.Environment`).

    Responsible for generating and wrapping a subprocess for execution and monitoring. Inherits from
    :class:`~threading.Thread` to run in separate thread (non-blocking monitoring). Call
    :func:`~Job.start` to execute (:func:`~Job.run` considered low-level API).

    Args:
        command: Program and arguments (iterable).
        cwd: Working directory.
        stdout: Callback, for each line of stdout.
        stderr: Callback, for each line of stderr.
        callback: Callback, for when task completes.
        keepends: Keep line breaks for individual lines?

    .. note:: Callback functions (arguments *stdout*, *stderr* and *callback*) must be threadsafe.
    """

    def __init__(self, command, cwd=None, stdout=None, stderr=None, callback=None, keepends=False):
        threading.Thread.__init__(self)

        self.init = Status()  #: True if job has started. :class:`Status` instance.
        self.done = Status()  #: True if job is done. :class:`Status` instance.

        self.command = tuple(str(x) for x in command)
        self.working_dir = cwd if cwd else None
        self.keepends = keepends

        self._callback = dict(stdout=stdout, stderr=stderr, done=callback)
        self._queue = dict(output=queue.Queue(), stdout=queue.Queue(), stderr=queue.Queue())
        self._history = dict(output=[], stdout=[], stderr=[])

    def run(self):
        """Starts execution *in current thread*.

        .. note:: You probably want :func:`~Job.start`.
        """

        if os.name == 'nt':
            asyncio.set_event_loop_policy(asyncio. WindowsProactorEventLoopPolicy())
        try:
            self.loop = asyncio.get_event_loop()
        except RuntimeError:
            self.loop = asyncio.new_event_loop()
            asyncio.set_event_loop(self.loop)
        if os.name == 'nt':
            assert isinstance(self.loop, asyncio.ProactorEventLoop), 'bad event loop for Windows'

        returncode = self.loop.run_until_complete(self._main())

        self._debug('Finishing job')
        for q in self._queue.values():
            q.put(None)
        return returncode

    async def wait(self, timeout=None, poll=1):
        """Awaits job completion.

        Arguments:
            timeout: Maximum wait time in seconds.
            poll: Wait between each status check of job.

        Raises:
            TimeoutError: Job did not complete in *timeout* seconds.

        Use ``None`` for *timeout* to wait forever. Polling impossible to avoid with Python's GIL
        (don't event try)!
        """

        if (timeout and timeout < 0) or (poll < 0):
            raise ValueError("arguments (timeout, poll) not positive numbers (seconds)")

        def fmt_sec(seconds):
            if not seconds:
                return 'infinite'
            elif seconds < 1:
                return '%.3f msec' % round(seconds*1000, 3)
            elif seconds < 60:
                return '%.3f sec' % round(seconds, 3)
            else:
                m, s = divmod(seconds, 60)
                h, m = divmod(m, 60)
                return '%d:%02d:%02d' % (h, m, s)

        start_time = time.time()
        self._debug('Waiting (timeout: %s) on job (poll: %s)', fmt_sec(timeout), fmt_sec(poll))
        while not self.done:
            seconds = time.time() - start_time
            if (timeout is not None) and (seconds > timeout):
                msg = 'Timed out (%s > %s) while waiting on job' % (fmt_sec(seconds),
                                                                    fmt_sec(timeout))
                self._log(logging.ERROR, '%s, terminating %r' % (msg, self.proc))
                self.proc.terminate()
                raise TimeoutError(msg)
            await asyncio.sleep(poll)

        self._info('Waited %s on job, now joining thread', fmt_sec(time.time() - start_time))
        if self.is_alive():
            self.join()

    @property
    def output(self):
        """Current, non-linesplit stdout/stderr (mixed) history."""
        if not self.proc:
            return None
        with threading.Lock():
            return ''.join(self._history['output'])

    @property
    def stdout(self):
        """Current, non-linesplit stdout history."""
        if not self.proc:
            return None
        with threading.Lock():
            return ''.join(self._history['stdout'])

    @property
    def stderr(self):
        """Current, non-linesplit stderr history."""
        if not self.proc:
            return None
        with threading.Lock():
            return ''.join(self._history['stderr'])

    @property
    def rc(self):
        """Return code of subprocess."""
        if not self.proc:
            return None
        with threading.Lock():
            return self.proc.returncode

    @property
    def pid(self):
        """PID of subprocess."""
        if not self.proc:
            return None
        with threading.Lock():
            return self.proc.pid

    @property
    def proc(self):
        """Started subprocess. :class:`Process` object."""
        with threading.Lock():
            try:
                return self._proc
            except AttributeError:
                return None

    @proc.setter
    def proc(self, value):
        with threading.Lock():
            self._proc = value

    def iter_stream(self, stream_type, timeout=0, keepends=None):
        """Line-for-line generator for :attr:`~Job.stdout`, :attr:`~Job.stderr` or
        :attr:`~Job.output` of non-processed lines in queue.

        Arguments:
            stream_type: *stdout*, *stderr* or *output* (mixed stream)
            timeout: Time to wait for new lines/process completion. Non-blocking if 0.
            keepends: Keep line breaks?

        .. todo:: Needs reference counting for multiple instances.
        """

        keepends = self.keepends if keepends is None else keepends
        get_line = functools.partial(self._queue[stream_type].get, timeout != 0, timeout)
        while True:
            try:
                line = get_line()
            except queue.Empty:
                return
            if line is None:
                return
            if keepends:
                yield line
            else:
                yield ''.join(line.splitlines())

    def _warn(self, msg, *args):
        self._log(logging.WARNING, msg % args)

    def _info(self, msg, *args):
        self._log(logging.INFO, msg % args)

    def _debug(self, msg, *args):
        self._log(logging.DEBUG, msg % args)

    def _log(self, level, msg):
        with threading.Lock():
            logging.getLogger(__name__).log(level, '%s: %r on <Thread %s>', msg,
                                            self, threading.current_thread().getName())

    async def _main(self):
        """Main coroutine.

        Subprocess running/scheduling. Only called internally.

        See `reference <https://stackoverflow.com/a/20697159>`_ of implementation.
        """

        if self.working_dir:
            working_dir = str(self.working_dir)
        else:
            working_dir = None

        try:
            self._debug('Starting subprocess %r', self.command)
            self.proc = await asyncio.create_subprocess_exec(*self.command, stdout=PIPE,
                                                             stderr=PIPE, cwd=working_dir)
        except Exception as exc:
            self._warn('Job failed to start! Process raised <%r>', exc)
            self.init.set()
            self.done.set()
            raise exc
        else:
            self._info('Job %r started', self.proc)
            self.init.set()

        await asyncio.gather(self._await_stream_lines(self.proc.stdout, 'stdout'),
                             self._await_stream_lines(self.proc.stderr, 'stderr'))
        self._debug('%r pipes closed', self.proc)

        return_code = await self.proc.wait()
        if return_code == 0:
            self._info('Job %r exited normally', self.proc)
        else:
            self._warn('Job %r error exited (rc=%d)', self.proc, return_code)

        if self._callback['done']:
            self._debug('Scheduling callback (to %r)', self._callback['done'])
            callback = functools.partial(self._callback['done'], self)
            self.loop.call_soon_threadsafe(self.loop.run_in_executor, None, callback)
            # await loop.run_in_executor(None, callback)
            # asyncio.run_coroutine_threadsafe(callback, loop=self.loop)

        self.done.set()
        return return_code

    def _stream_handle(self, line, name):
        """Handler of all stdout/stderr lines.

        Put in queue, history and call external handler (if any).
        """

        line = line.decode(locale.getpreferredencoding(False))
        with threading.Lock():
            self._history['output'] += [line]
            self._history[name] += [line]

        if not self.keepends:
            line = ''.join(line.splitlines())
        self._queue['output'].put(line)
        self._queue[name].put(line)

        if self._callback[name]:
            callback = functools.partial(self._callback[name], line)
            asyncio.get_event_loop().call_soon_threadsafe(callback)

    async def _await_stream_lines(self, reader, name):
        """Awaits new line in *reader*, calling handler threadsafely."""
        while True:
            line = await reader.readline()
            if not line:
                break
            self._debug('Read %r [%s]', line, name.upper())
            callback = functools.partial(self._stream_handle, line, name)
            asyncio.get_event_loop().call_soon_threadsafe(callback)
        self._debug('Read EOF [%s]', name.upper())

    def __repr__(self):
        with threading.Lock():
            name = '%s (%s)' % (hex(id(self)), self.getName())
            if self.done.is_set():
                return '<done %s %s>' % (self.__class__.__name__, name)
            elif self.init.is_set():
                return '<busy %s %s, %s>' % (self.__class__.__name__, name, self.proc)
            else:
                return '<new %s %s>' % (self.__class__.__name__, name)



import subprocess
from pathlib import Path

from toil.job import Job

from pharmpy import Model


class InputModel:
    """A wrapper for a model that is input to a specific workflow

       Handles all files needed for the model and the model object
    """
    def __init__(self, workflow, model):
        self.model = model
        self.files = import_files(workflow, [model.input.path])


class CommandLineJob(Job):
    def copy_file(self, file_store, file_id):
        path = file_store.readGlobalFile(file_id, userPath=Path(file_id).name)
        return Path(path)

    def call(self, command, file_store):
        output = subprocess.check_output(command).decode()
        file_store.logToMaster(output)


class GatherFiles(Job):
    """Gather files after a batch job
    """
    def __init__(self, rvs):
        super().__init__()
        self.rvs = rvs

    def run(self, file_store):
        return self.rvs


class BatchModelfit(Job):
    """Runs a list of models

       root -> [NonmemJob ...] -> GatherFiles
    """
    def __init__(self, models):
        super().__init__()
        self.input_models = models

    def run(self, file_store):
        rvs = []
        for input_model in self.input_models:
            job = NonmemJob(input_model.model, input_model.files)
            child = self.addChild(job)
            rvs.append(child.rv())

        return self.addFollowOn(GatherFiles(rvs)).rv()


class NonmemJob(CommandLineJob):
    """One nmfe execution
    """
    def __init__(self, model, additional_files):
        super().__init__()
        self.model = model
        self.additional_files = additional_files

    def run(self, file_store):
        path = self.model.write()
        for file_id in self.additional_files:
            self.copy_file(file_store, file_id)
        self.call(['nmfe74', path.name, path.with_suffix('.lst').name], file_store)
        model = Model(path)        # New model for now. Paths need updating...
        model.modelfit_results.parameter_estimates      # To read it in
        result_files = \
            [file_store.writeGlobalFile(str(path)) for path in model.modelfit_results.tool_files]
        model.modelfit_results.tool_files = result_files
        return model


def import_file(workflow, path):
    path = Path(path).resolve()
    url = "file://" + str(path)
    file_id = workflow.importFile(url)
    return file_id


def import_files(workflow, paths):
    file_ids = [import_file(workflow, path) for path in paths]
    return file_ids


def export_files(workflow, file_ids):
    dir_path = Path('.').resolve()
    for file_id in file_ids:
        path = dir_path / Path(file_id).name
        url = 'file://' + str(path)
        workflow.exportFile(file_id, url)



import pandas as pd
import pytest

import pharmpy.data as data


@pytest.fixture
def df():
    df = data.PharmDataFrame({'ID': [1, 1, 2, 2], 'DV': [0.1, 0.2, 0.5, 0.6],
                              'WGT': [70, 72, 75, 75], 'HGT': [185, 185, 160, 160]})
    df.pharmpy.column_type[('ID', 'DV', 'WGT', 'HGT')] = [
        data.ColumnType.ID, data.ColumnType.DV, data.ColumnType.COVARIATE,
        data.ColumnType.COVARIATE]
    return df


def test_column_type():
    col_type = data.ColumnType.ID
    assert col_type == data.ColumnType.ID
    assert 'ID' in repr(col_type)
    assert col_type.max_one
    ct2 = data.ColumnType.UNKNOWN
    assert ct2 == data.ColumnType.UNKNOWN
    assert 'UNKNOWN' in repr(ct2)
    assert not ct2.max_one


def test_data_frame():
    df = data.PharmDataFrame({'ID': [1, 1, 2, 2], 'DV': [0.1, 0.2, 0.5, 0.6]})
    assert list(df.columns) == ['ID', 'DV']


def test_accessor_get_set_column_type():
    df = data.PharmDataFrame({'ID': [1, 1, 2, 2], 'DV': [0.1, 0.2, 0.5, 0.6],
                              'WGT': [70, 70, 75, 75], 'HGT': [185, 185, 160, 160]})
    assert df.pharmpy.column_type['ID'] == data.ColumnType.UNKNOWN
    assert df.pharmpy.column_type[('ID', 'DV')] == \
        [data.ColumnType.UNKNOWN, data.ColumnType.UNKNOWN]

    df.pharmpy.column_type['ID'] = data.ColumnType.ID
    df.pharmpy.column_type['DV'] = data.ColumnType.DV
    with pytest.raises(KeyError):       # Max one id column
        df.pharmpy.column_type['DV'] = data.ColumnType.ID
    with pytest.raises(KeyError):
        df.pharmpy.column_type['NOEXISTS'] = data.ColumnType.COVARIATE
    with pytest.raises(ValueError):
        df.pharmpy.column_type[('ID', 'DV')] = \
            [data.ColumnType.COVARIATE, data.ColumnType.COVARIATE, data.ColumnType.COVARIATE]
    with pytest.raises(KeyError):
        df.pharmpy.column_type['NOAVAIL']
    assert df.pharmpy.column_type['ID'] == data.ColumnType.ID
    assert df.pharmpy.column_type['DV'] == data.ColumnType.DV
    assert df.pharmpy.column_type[['ID', 'DV']] == [data.ColumnType.ID, data.ColumnType.DV]
    df.pharmpy.column_type[['HGT', 'WGT']] = data.ColumnType.COVARIATE
    assert df.pharmpy.labels_by_type[data.ColumnType.COVARIATE] == ['WGT', 'HGT']
    df2 = df.copy()
    assert df2.pharmpy.column_type['ID'] == data.ColumnType.ID
    assert df2.pharmpy.column_type['DV'] == data.ColumnType.DV

    assert df2.pharmpy.labels_by_type[data.ColumnType.ID] == ['ID']
    assert df2.pharmpy.labels_by_type[[data.ColumnType.ID]] == ['ID']
    assert df2.pharmpy.labels_by_type[[]] == []
    assert df2.pharmpy.labels_by_type[data.ColumnType.DV] == ['DV']
    assert df2.pharmpy.labels_by_type[[data.ColumnType.ID, data.ColumnType.DV]] == ['ID', 'DV']

    assert df2.pharmpy.id_label == 'ID'

    df2.pharmpy.column_type[['ID', 'DV']] = (data.ColumnType.COVARIATE, data.ColumnType.IDV)
    assert df2.pharmpy.column_type[['ID', 'DV']] == [data.ColumnType.COVARIATE, data.ColumnType.IDV]
    assert df.pharmpy.column_type['ID'] == data.ColumnType.ID

    df3 = data.PharmDataFrame({1: [1, 1, 2, 2], 2: [0.1, 0.2, 0.5, 0.6],
                               3: [70, 70, 75, 75], 4: [185, 185, 160, 160]})
    assert df3.pharmpy.column_type[1] == data.ColumnType.UNKNOWN
    df3.pharmpy.column_type[[3, 4]] = data.ColumnType.COVARIATE
    df3.pharmpy.column_type[1] = data.ColumnType.ID
    assert df3.pharmpy.column_type[[1, 2, 3, 4]] == \
        [data.ColumnType.ID, data.ColumnType.UNKNOWN,
         data.ColumnType.COVARIATE, data.ColumnType.COVARIATE]


def test_time_varying_covariates(df):
    assert df.pharmpy.time_varying_covariates == ['WGT']
    df_untyped = data.PharmDataFrame({'ID': [1, 1, 2, 2], 'DV': [0.1, 0.2, 0.5, 0.6]})
    assert df_untyped.pharmpy.time_varying_covariates == []


def test_covariate_baselines(df):
    correct_baselines = pd.DataFrame(
        {'WGT': [70, 75], 'HGT': [185, 160]},
        index=pd.Int64Index([1, 2], name='ID')
    )
    pd.testing.assert_frame_equal(df.pharmpy.covariate_baselines, correct_baselines)


def test_write(fs, df):
    df.pharmpy.write_csv(path="my.csv")
    with open("my.csv", "r") as fh:
        contents = fh.read()
    print(contents)
    assert contents == "ID,DV,WGT,HGT\n1,0.1,70,185\n1,0.2,72,185\n2,0.5,75,160\n2,0.6,75,160\n"
